# %%
from typing import List, Union
import os
import requests
import torch
import cv2
import numpy as np
import matplotlib.pyplot as plt
from torchvision import transforms
from PIL import Image
import yaml
import pandas as pd
import pickle
from fundus_image_toolbox.utilities import ImageTorchUtils as Img

from .SuperRetina import (
    pre_processing,
    simple_nms,
    remove_borders,
    sample_keypoint_desc,
    SuperRetina,
)

WEIGHT_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "SuperRetina/save/SuperRetina.pth"
)
IMG_SIZE = 512
DEFAULT_CONFIG = {
    "device": "cuda:0",
    "use_matching_trick": True,
    "nms_size": 1,
    "nms_thresh": 0.0005,
    "knn_thresh": 0.85,
}


def wget(link, target):
    # platform independent wget alternative
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36"
    }
    r = requests.get(link, headers=headers, stream=True)
    downloaded_file = open(target, "wb")

    for chunk in r.iter_content(chunk_size=8192):
        if chunk:
            downloaded_file.write(chunk)


def download_weights(url="https://zenodo.org/records/11241985/files/SuperRetina.pth"):
    os.makedirs(os.path.dirname(WEIGHT_PATH), exist_ok=True)
    print("Downloading weights...")
    wget(url, WEIGHT_PATH)
    print("Done")


def get_mask(filename, masks: pd.DataFrame = None, show=True, binarize=True):
    directory = os.path.dirname(filename)
    if masks is None:
        p = os.path.join(directory, "masks.pkl")
        masks = pickle.load(open(p, "rb"))

    row = masks.loc[masks["image"] == os.path.basename(filename)]
    mask = row["mask"].values[0]
    size = row["size"].values[0]

    if binarize:
        thresh = 0.2
        mask = [1 if x > thresh else 0 for x in mask]

    mask = np.array(mask).reshape(size, size)

    if show:
        plt.imshow(mask)
        plt.show()
        plt.close()

    assert np.sum(mask) > 0

    return mask


# %%
def any_to_image_array(image):
    if isinstance(image, str):
        image = cv2.imread(image)
    elif isinstance(image, Image.Image):
        # To BGR
        image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
        image = np.array(image)
    elif isinstance(image, np.ndarray):
        # To BGR
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    return image


def any_to_greyscale(image):
    image = any_to_image_array(image)
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    # Cast to Uint8
    image = image.astype(np.uint8)
    return image


# %%
def enhance(image):
    image = any_to_greyscale(image)
    black_mask = image < 10

    # Replace black pixels with mean of the image
    image[black_mask] = image[~black_mask].mean()

    # Normalize
    image = (image - image.min()) / (image.max() - image.min())

    # Enhance images using clahe
    image = pre_processing(image)
    image = (image * 255).astype(np.uint8)

    # Add in the mask again
    image[black_mask == 1] = 0

    return image


# %%
def transform(image: Image, model_image_height, model_image_width):
    compose = transforms.Compose(
        [
            # Resize the image to the model input size at smaller edge length
            transforms.Resize(model_image_height),
            # Center crop the image
            transforms.CenterCrop((model_image_height, model_image_width)),
            transforms.ToTensor(),
        ]
    )

    return compose(image)


# %%
def get_config(config=Union[dict, str, None]):
    """Get the configuration for inference. If config is None, use the default config. Adds the
    path to the model weights, if needed, as well as the image size for the model
    (Default: 512x512). The latter is independent of the image output size, so you can leave it at
    the default.

    Args:
        config (dict or str, optional): The configuration. Defaults to None. Options:
            - None
            - str: path to yaml file
            - dict: configuration dictionary

    Returns:
        dict: The configuration dictionary
    """
    if config is None:
        config = DEFAULT_CONFIG
    elif isinstance(config, str):
        if os.path.exists(config):
            with open(config) as f:
                config = yaml.safe_load(f)
        else:
            raise FileNotFoundError("Config file not found.")
    elif isinstance(config, dict):
        pass
    else:
        raise ValueError("Config must be either a path to a yaml file or a dictionary")

    if "PREDICT" in config:
        config = config["PREDICT"]

    if not hasattr(config, "model_save_path"):
        config["model_save_path"] = WEIGHT_PATH
    if not hasattr(config, "model_image_height"):
        config["model_image_height"] = IMG_SIZE
    if not hasattr(config, "model_image_width"):
        config["model_image_width"] = IMG_SIZE

    return config


# %%
def get_device(config: dict):
    device = (
        torch.device(config["device"])
        if torch.cuda.is_available()
        else torch.device("cpu")
    )
    return device


# %%
def load_model(config: dict = None):
    """Load the SuperRetina model and the knn matcher for matching keypoints. Returns the model and the matcher.

    Args:
        config (dict): The configuration dictionary
    """
    if config is None:
        config = DEFAULT_CONFIG

    config = get_config(config)

    device = get_device(config)

    if not os.path.isfile(config["model_save_path"]):
        download_weights()

    model = SuperRetina().to(device)
    model.load_state_dict(
        torch.load(config["model_save_path"], map_location=device)["net"]
    )
    model.eval()

    knn_matcher = cv2.BFMatcher(cv2.NORM_L2)

    return model, knn_matcher


# %%
def predict_points(fixed_tensor, moving_tensor, config, model):
    inputs = torch.cat((moving_tensor.unsqueeze(0), fixed_tensor.unsqueeze(0)))
    inputs = inputs.to(get_device(config))

    with torch.no_grad():
        detector_pred, descriptor_pred = model(inputs)

    scores = simple_nms(detector_pred, config["nms_size"])

    b, _, h, w = detector_pred.shape
    scores = scores.reshape(-1, h, w)

    keypoints = [torch.nonzero(s > config["nms_thresh"]) for s in scores]

    scores = [s[tuple(k.t())] for s, k in zip(scores, keypoints)]

    # Discard keypoints near the image borders
    keypoints, scores = list(
        zip(*[remove_borders(k, s, 4, h, w) for k, s in zip(keypoints, scores)])
    )

    keypoints = [torch.flip(k, [1]).float().data for k in keypoints]

    descriptors = [
        sample_keypoint_desc(k[None], d[None], 8)[0].cpu().data
        for k, d in zip(keypoints, descriptor_pred)
    ]
    keypoints = [k.cpu() for k in keypoints]

    return keypoints, descriptors


# %%
def get_match_visualization(imageA, imageB, kpsA, kpsB, matches, status):
    # Initialize the output visualization image
    (hA, wA) = imageA.shape[:2]
    (hB, wB) = imageB.shape[:2]
    vis = np.zeros((max(hA, hB), wA + wB, 3), dtype="uint8")
    if len(imageA.shape) == 2:
        imageA = cv2.cvtColor(imageA, cv2.COLOR_GRAY2RGB)
        imageB = cv2.cvtColor(imageB, cv2.COLOR_GRAY2RGB)

    vis[0:hA, 0:wA] = imageA
    vis[0:hB, wA:] = imageB

    # Loop over the matches
    for i, ((match, _), s) in enumerate(zip(matches, status)):
        trainIdx, queryIdx = match.trainIdx, match.queryIdx
        # Only process the match if the keypoint was successfully matched
        if s == 1:
            # Draw the match
            ptA = (int(kpsA[queryIdx].pt[0]), int(kpsA[queryIdx].pt[1]))
            ptB = (int(kpsB[trainIdx].pt[0]) + wA, int(kpsB[trainIdx].pt[1]))
            # Each line gets a color
            color = tuple(map(int, np.random.randint(0, 255, size=3)))
            cv2.line(vis, ptA, ptB, color, 2)

    return vis


# %%
def match(matcher, fixed_desc, moving_desc, config):
    """Match the keypoints using the knn matcher"""
    goodMatch = []
    status = []
    try:
        matches = matcher.knnMatch(moving_desc, fixed_desc, k=2)
        for m, n in matches:
            if m.distance < config["knn_thresh"] * n.distance:
                goodMatch.append(m)
                status.append(True)
            else:
                status.append(False)
    except Exception:
        pass

    return goodMatch, status, matches


def check_collapse(
    goodMatch, cv_kpts_fixed, cv_kpts_moving, round_to: int = 1, threshold: float = 0.65
):
    """Filter out matches where keypoints in source are projected to one point in target (and vice versa)"""
    # Get keypoints as list of tuples
    moving_keypoints = [cv_kpts_moving[m.queryIdx].pt for m in goodMatch]
    fixed_keypoints = [cv_kpts_fixed[m.trainIdx].pt for m in goodMatch]
    moving_keypoints = [
        (round(kp[0], round_to), round(kp[1], 2)) for kp in moving_keypoints
    ]
    fixed_keypoints = [
        (round(kp[0], round_to), round(kp[1], 2)) for kp in fixed_keypoints
    ]

    # Get unique keypoints
    unique_moving = len(list(set(moving_keypoints)))
    unique_fixed = len(list(set(fixed_keypoints)))
    ratio_moving = unique_moving / len(moving_keypoints)
    ratio_fixed = unique_fixed / len(fixed_keypoints)
    if ratio_moving < threshold or ratio_fixed < threshold:
        raise Exception(
            "Failed to align the two images: Too many points are projected to one!"
        )


# %%
def find_homography_and_align(
    fixed_image,
    moving_image,
    original_moving_image,
    cv_kpts_fixed,
    cv_kpts_moving,
    goodMatch,
    matches,
    status,
    show=False,
    show_mapping=False,
    save_to=None,
):
    # Check if too many points are projected to one
    check_collapse(goodMatch, cv_kpts_fixed, cv_kpts_moving)

    H_m = None
    inliers_num_rate = 0
    good = goodMatch.copy()
    if len(goodMatch) >= 4:
        src_pts = [cv_kpts_moving[m.queryIdx].pt for m in good]
        src_pts = np.float32(src_pts).reshape(-1, 1, 2)
        dst_pts = [cv_kpts_fixed[m.trainIdx].pt for m in good]
        dst_pts = np.float32(dst_pts).reshape(-1, 1, 2)

        H_m, mask = cv2.findHomography(src_pts, dst_pts, cv2.LMEDS)

        good = np.array(good)[mask.ravel() == 1]
        status = np.array(status)
        temp = status[status == True]
        temp[mask.ravel() == 0] = False
        status[status == True] = temp
        inliers_num_rate = mask.sum() / len(mask.ravel())

    if show_mapping:
        query_np = np.array([kp.pt for kp in cv_kpts_moving])
        refer_np = np.array([kp.pt for kp in cv_kpts_fixed])
        refer_np[:, 0] += moving_image.shape[1]
        matched_image = get_match_visualization(
            moving_image, fixed_image, cv_kpts_moving, cv_kpts_fixed, matches, status
        )
        plt.figure(dpi=300)
        plt.scatter(query_np[:, 0], query_np[:, 1], s=1, c="r")
        plt.scatter(refer_np[:, 0], refer_np[:, 1], s=1, c="r")
        plt.axis("off")
        plt.title("Match Result, #goodMatch: {}".format(len(good)))
        plt.imshow(cv2.cvtColor(matched_image, cv2.COLOR_BGR2RGB))
        plt.show()
        plt.close()

        print("The rate of inliers: {:.3f}%".format(inliers_num_rate * 100))

    # Get dims of fixed image
    image_height, image_width = fixed_image.shape[:2]

    if H_m is not None:
        h, w = image_height, image_width
        original_moving_align = cv2.warpPerspective(
            original_moving_image,
            H_m,
            (w, h),
            borderMode=cv2.BORDER_CONSTANT,
            borderValue=(0),
        )
        moving_align = original_moving_align.copy()

        merged = np.zeros((h, w, 3), dtype=np.uint8)

        if len(moving_align.shape) == 3:
            moving_align = enhance(moving_align)
            original_moving_align = cv2.cvtColor(
                original_moving_align, cv2.COLOR_BGR2RGB
            )

        if len(fixed_image.shape) == 3:
            fixed_gray = cv2.cvtColor(fixed_image, cv2.COLOR_BGR2GRAY)
        else:
            fixed_gray = fixed_image
        merged[:, :, 0] = moving_align
        merged[:, :, 1] = fixed_gray

    else:
        raise Exception("Failed to align the two images!")

    def fig_grid():
        # Create 1x4 grid and return the axes
        fig, ax = plt.subplots(1, 4, figsize=(20, 5))
        for a in ax:
            a.axis("off")
        ax[0].imshow(fixed_gray, "gray")
        ax[0].set_title("fixed")
        ax[1].imshow(moving_image, "gray")
        ax[1].set_title("moving")
        ax[2].imshow(merged)
        ax[2].set_title("merged")
        ax[3].imshow(original_moving_align)
        ax[3].set_title("out")

        return fig, ax

    if H_m is not None and show:
        fig, ax = fig_grid()
        plt.show()

    if save_to is not None and H_m is not None:
        fig, ax = fig_grid()
        plt.savefig(save_to)
        plt.close()

    return original_moving_align


# %%
def register_one(
    fixed_image: Union[str, np.ndarray, Image.Image],
    moving_image: Union[str, np.ndarray, Image.Image],
    show: bool = True,
    show_mapping: bool = False,
    save_to: str = None,
    config: dict = DEFAULT_CONFIG,
    model=None,
    matcher=None,
):
    """Register two images using SuperRetina and homography transformation.

    Args:
        fixed_image (Union[str, np.array, Image.Image]): The fixed image.
        moving_image (Union[str, np.array, Image.Image]): The moving image to register to the fixed.
        show (bool, optional): Show the images. Defaults to True.
        show_mapping (bool, optional): Show the mapping of SuperPoints. Defaults to False.
        save_to (str, optional): Save the images to this path. Defaults to None.
        config (dict, optional): The configuration dict for inference. Defaults to DEFAULT_CONFIG.
        model (optional): The SuperRetina model. Defaults to None.
        matcher (optional): The matcher for matching keypoints. Defaults to None.

    Returns:
        [np.array]: The aligned moving image.
    """

    if save_to is not None:
        if not os.path.exists(save_to):
            os.makedirs(save_to)
        if isinstance(fixed_image, str) and isinstance(moving_image, str):
            save_to = os.path.join(
                save_to,
                os.path.basename(fixed_image).__str__()
                + "_"
                + os.path.basename(moving_image).__str__(),
            )
        else:
            file_count = len(os.listdir(save_to))
            save_to = os.path.join(save_to, "aligned_" + str(file_count))

    original_moving_image = any_to_image_array(moving_image)

    fixed_image = enhance(fixed_image)
    moving_image = enhance(moving_image)

    config = get_config(config)

    fixed_tensor = transform(
        Image.fromarray(fixed_image),
        config["model_image_height"],
        config["model_image_width"],
    )
    moving_tensor = transform(
        Image.fromarray(moving_image),
        config["model_image_height"],
        config["model_image_width"],
    )

    image_height, image_width = fixed_image.shape

    if model is None or matcher is None:
        model, matcher = load_model(config)

    keypoints, descriptors = predict_points(fixed_tensor, moving_tensor, config, model)

    fixed_keypoints, moving_keypoints = keypoints[1], keypoints[0]
    fixed_desc, moving_desc = (
        descriptors[1].permute(1, 0).numpy(),
        descriptors[0].permute(1, 0).numpy(),
    )

    # Mapping keypoints to scaled keypoints
    cv_kpts_moving = [
        cv2.KeyPoint(
            int(i[0] / config["model_image_width"] * image_width),
            int(i[1] / config["model_image_height"] * image_height),
            30,
        )  # 30 is keypoints size, which can be ignored
        for i in moving_keypoints
    ]
    cv_kpts_fixed = [
        cv2.KeyPoint(
            int(i[0] / config["model_image_width"] * image_width),
            int(i[1] / config["model_image_height"] * image_height),
            30,
        )
        for i in fixed_keypoints
    ]

    goodMatch, status, matches = match(matcher, fixed_desc, moving_desc, config)

    moving_aligned = find_homography_and_align(
        fixed_image,
        moving_image,
        original_moving_image,
        cv_kpts_fixed,
        cv_kpts_moving,
        goodMatch,
        matches,
        status,
        show=show,
        show_mapping=show_mapping,
        save_to=save_to,
    )

    return moving_aligned


def register(
    fixed_image: Union[
        str,
        List[str],
        List[np.ndarray],
        List[torch.Tensor],
        List[Image.Image],
        np.ndarray,
        torch.Tensor,
        Image.Image,
    ],
    moving_image: Union[
        str,
        List[str],
        List[np.ndarray],
        List[torch.Tensor],
        List[Image.Image],
        np.ndarray,
        torch.Tensor,
        Image.Image,
    ],
    show: bool = True,
    show_mapping: bool = False,
    save_to: str = None,
    config: dict = DEFAULT_CONFIG,
    model=None,
    matcher=None,
) -> Union[np.ndarray, List[np.ndarray]]:
    """Register a pair or pairs of images from batches using SuperRetina and homography transformation.

    Args:
        fixed_image (Union[torch.Tensor, list, str, np.array, Image.Image]): The fixed image(s).
        moving_image (Union[torch.Tensor, list, str, np.array, Image.Image]): The moving image(s) to register to the fixed.
        show (bool, optional): Show the images. Defaults to True.
        show_mapping (bool, optional): Show the mapping of SuperPoints. Defaults to False.
        save_to (str, optional): Save the images to this path. Defaults to None.
        config (dict, optional): The configuration dict for inference. Defaults to DEFAULT_CONFIG.
        model (optional): The SuperRetina model. Defaults to None.
        matcher (optional): The matcher for matching keypoints. Defaults to None.

    Returns:
        list of np.ndarrays or np.ndarray: The aligned moving image(s).
    """

    # To batch of numpy images
    fixed_image = Img(fixed_image).to_tensor().squeeze().to_batch().to_numpy().img
    moving_image = Img(moving_image).to_tensor().squeeze().to_batch().to_numpy().img

    if len(fixed_image) != len(moving_image):
        raise ValueError("The number of fixed and moving images must be the same.")

    moving_aligned = []
    for fixed, moving in zip(fixed_image, moving_image):
        moving_aligned.append(
            register_one(
                fixed, moving, show, show_mapping, save_to, config, model, matcher
            )
        )

    if len(moving_aligned) == 1:
        return moving_aligned[0]
    return moving_aligned


if __name__ == "__main__":
    fixed_image = "./example1.jpg"
    moving_image = "./example2.jpg"

    moving_aligned = register(fixed_image, moving_image, show=True)
    plt.imshow(moving_aligned, "gray")



import unittest
import os
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import cv2
import torch
from torchvision.transforms.functional import to_tensor
from fundus_image_toolbox.utilities import ImageTorchUtils as Img

DIR = os.path.join(os.path.dirname(__file__))

fundus1_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus1.jpg")
fundus2_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus2.jpg")

fundus1_plt = plt.imread(fundus1_path)
fundus1_img = Image.open(fundus1_path)
fundus1_cv2 = cv2.imread(fundus1_path)


class TestImageTorchUtils(unittest.TestCase):
    def test_to_tensor(self):
        # Test the to_tensor function
        imgs_rgb = [fundus1_path, fundus1_plt, fundus1_img]

        # Convert to rgb tensor from rgb image
        outs = [Img(img).to_tensor().img for img in imgs_rgb]

        # Convert to rgb tensor from bgr image
        bgr_out = Img(fundus1_cv2).to_tensor(from_cspace="bgr").img
        # Alternatively, use a sequence of calls:
        bgr_out = (
            Img(fundus1_cv2)
            .to_tensor()
            .to_cspace(from_cspace="bgr", to_cspace="rgb")
            .img
        )
        outs.append(bgr_out)

        assert all([torch.equal(outs[0], out) for out in outs[1:]])

    def test_to_cspace(self):
        # Test the to_cspace function
        imgs_rgb = [fundus1_path, fundus1_plt, fundus1_img]
        img_bgr = fundus1_cv2

        # Convert from rgb to bgr
        outs = [
            Img(img).to_tensor().to_cspace(from_cspace="rgb", to_cspace="bgr").img
            for img in imgs_rgb
        ]

        assert all([torch.equal(outs[0], out) for out in outs[1:]])
        assert torch.equal(Img(img_bgr).to_tensor().img, outs[0])

        # Convert to grey scale
        outs = [
            Img(img).to_tensor().to_cspace(from_cspace="rgb", to_cspace="grey").img
            for img in imgs_rgb
        ]

        assert all([torch.equal(outs[0], out) for out in outs[1:]])
        assert torch.equal(
            Img(img_bgr).to_tensor().to_cspace(from_cspace="bgr", to_cspace="grey").img,
            outs[0],
        )

    def test_set_channel_dim(self):
        # Test set_channel_dim function
        imgs = [fundus1_path, fundus1_plt, fundus1_img, fundus1_cv2]

        # Set channel dimension to the last dimension
        outs = [Img(img).to_tensor().set_channel_dim(-1).img for img in imgs]

        # Check if the channel dimension is the last dimension
        assert all([out.shape[-1] == 3 for out in outs])

    def test_to_pil(self):
        # Test to_pil function
        imgs = [fundus1_path, fundus1_plt, fundus1_img, fundus1_cv2]

        # Convert to pil image
        outs = [Img(img).to_tensor().to_pil().img for img in imgs]

        # Check if the output is a pil image
        assert all([isinstance(out, Image.Image) for out in outs])

    def test_to_numpy(self):
        # Test to_numpy functions
        imgs = [fundus1_path, fundus1_plt, fundus1_img, fundus1_cv2]

        # Convert to numpy array
        outs = [Img(img).to_tensor().to_numpy().img for img in imgs]

        # Check if the output is a numpy array
        assert all([isinstance(out, np.ndarray) for out in outs])
        assert all([out.dtype == np.uint8 for out in outs])

        # Alternatively, chain the to_numpy and to_unit8 functions
        outs = [Img(img).to_tensor().to_numpy().to_uint8().img for img in imgs]

        # Check if the output is a numpy array
        assert all([isinstance(out, np.ndarray) for out in outs])
        assert all([out.dtype == np.uint8 for out in outs])

        # Convert to float32 numpy array
        outs = [Img(img).to_tensor().to_numpy("float32").img for img in imgs]

        # Check if the output is a numpy array
        assert all([isinstance(out, np.ndarray) for out in outs])
        assert all([out.dtype == np.float32 for out in outs])

        # Alternatively, chain the to_numpy and to_float32 functions
        outs = [Img(img).to_tensor().to_numpy().to_float32().img for img in imgs]

        # Check if the output is a numpy array
        assert all([isinstance(out, np.ndarray) for out in outs])
        assert all([out.dtype == np.float32 for out in outs])

    def test_to_batch(self):
        # Test the to_batch function

        # Test objects
        list_fundus_plt = [fundus1_plt, fundus1_plt]
        list_fundus_img = [fundus1_img, fundus1_img]
        list_fundus_cv2 = [fundus1_cv2, fundus1_cv2]

        batch_fundus = torch.stack([to_tensor(fundus1_img), to_tensor(fundus1_img)])

        ndarray_fundus_plt = np.array(list_fundus_plt)
        ndarray_fundus_img = np.array(list_fundus_img)
        ndarray_fundus_cv2 = np.array(list_fundus_cv2)

        # Apply the to_batch function
        outs = [
            Img(fundus).to_batch().img
            for fundus in [list_fundus_plt, list_fundus_img, list_fundus_cv2]
        ]
        outs += [
            Img(fundus).to_batch().img
            for fundus in [ndarray_fundus_plt, ndarray_fundus_img, ndarray_fundus_cv2]
        ]

        # Check if the output is a tensor
        for out in outs:
            assert isinstance(out, torch.Tensor)
            assert (
                out.shape[0] == len(batch_fundus)
                and out.shape[1] == 3
                and isinstance(out, torch.Tensor)
            )

        # Single image to batch
        out = Img(fundus1_img).to_batch().img
        assert len(out.shape) == 4 and isinstance(out, torch.Tensor)

        # Single path to batch
        out = Img(fundus1_path).to_batch().img
        assert len(out.shape) == 4 and isinstance(out, torch.Tensor)

    def test_combi(self):
        # Test the to_cspace and set_channel_dim function
        out = (
            Img(fundus1_plt)
            .to_tensor()
            .to_cspace(from_cspace="rgb", to_cspace="gray")
            .set_channel_dim(-1)
            .img
        )
        assert out.shape[-1] == 1


if __name__ == "__main__":
    unittest.main()



import unittest
import os
import numpy as np
import torch
from PIL import Image
from torchvision.transforms import Compose, ToTensor, Resize, CenterCrop
from fundus_image_toolbox.fovea_od_localization import ODFoveaModel, load_fovea_od_model

DIR = os.path.join(os.path.dirname(__file__))
fundus1_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus1.jpg")
DEVICE = "cpu"


class TestFoveaODModel(unittest.TestCase):
    def setUp(self):
        self.image = Image.open(fundus1_path)
        transform = Compose([Resize(350, antialias=True), CenterCrop(350), ToTensor()])
        self.image_batch = torch.stack([transform(self.image) for _ in range(2)])
        self.device = DEVICE

    def test_load_fovea_od_model(self):
        model, checkpoint_path = load_fovea_od_model(
            "default", device=self.device, return_test_dataloader=False
        )
        self.assertIsInstance(model, ODFoveaModel)
        self.assertEqual(self.device, str(model.device))

        fovea_x, fovea_y, od_x, od_y = model.predict(self.image)
        self.assertIsInstance(float(fovea_x), float)
        self.assertIsInstance(float(fovea_y), float)
        self.assertIsInstance(float(od_x), float)
        self.assertIsInstance(float(od_y), float)

        labels = model.predict(self.image_batch)
        self.assertIsInstance(labels, list)
        self.assertEqual(len(labels), 2)
        self.assertIsInstance(labels[0], np.ndarray)
        self.assertEqual(len(labels[0]), 4)
        self.assertIsInstance(float(labels[0][0]), float)


if __name__ == "__main__":
    unittest.main()



from .dataset_multi import ODFoveaLoader
from .model_multi import ODFoveaModel, plot_coordinates, plot_input
from .train_multi import evaluate_ensemble, get_test_dataloader, Parser
from .train_multi import load_model as load_fovea_od_model
from .default import MODELS_DIR, DEFAULT_CSV_PATH, DEFAULT_MODEL, DEFAULT_CONFIG


import os
from setuptools import setup, find_packages

project = os.path.basename(os.path.dirname(os.path.realpath(__file__)))
# about = __import__(project).about

with open("Readme.md", "r", encoding='utf8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name=project,
    version="0.0.1",
    author="Julius Gervelmeyer et al.",
    author_email="Julius.Gervelmeyer@uni-tuebingen.de",
    description=long_description.split('\n')[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url=None,
    packages=find_packages(),
    install_requires=required,
    package_dir={'': '.'},
    python_requires='>=3.9.0, <3.10',
)

def clone_repo():
    # Clone Segmentation Quality Control repo
    link = "https://github.com/berenslab/MIDL24-segmentation_quality_control"
    commit = "995a173c1c1735b620c14b2268ce3ca408c10ca1"
    branch = "main"
    target_dir = f"./{project}/segmentation"
    os.makedirs(target_dir, exist_ok=True)
    if not os.listdir(target_dir):
        pwd = os.getcwd()
        os.system(f"git clone {link} {target_dir}")
        os.system(f"cd {target_dir} && git checkout {branch} && git reset --hard {commit}")
        os.system(f"rm -rf {target_dir}/.git")
        os.chdir(pwd)
        print("Repo cloned successfully.")

        print("Adjusting imports...")
        add_dots_to_imports_in_folder(target_dir)
        print("Done.")

# Make imports relative
def add_dots_to_imports(file_path):
    with open(file_path, 'r') as file:
        lines = file.readlines()
    with open(file_path, 'w') as file:
        for line in lines:
            is_from_import = line.split()[0] == 'from' if len(line.split()) > 1 else False
            is_import = line.split()[0] == 'import' if len(line.split()) > 1 else False
            is_local_file = False
            is_local_folder = False
            
            if is_from_import or is_import:
                this_dir = os.path.dirname(file_path)
                par_dir = os.path.dirname(this_dir)
                fname = line.split()[1]
                if fname.startswith('.'):
                    file.write(line)
                    continue
                fname = fname.split('.')[0]
                # print(fname)

                # Check if is a local file in same directory
                if os.path.exists(os.path.join(this_dir, fname)):
                    print(os.path.join(this_dir, fname + '.py'))
                    is_local_file = True
                # Check if is a local folder in parent directory
                elif os.path.exists(os.path.join(par_dir, fname)):
                    print(os.path.join(par_dir, fname))
                    is_local_folder = True

            c = ".." if is_local_folder else "."
            if is_local_file or is_local_folder:
                terms = line.replace("import", ".").replace("from", "").replace("\n", "").split(".")
                terms = [t.replace(" ", "") for t in terms]
                parents = terms[:-1] if len(terms) > 1 else []
                imports = terms[-1].split(",")
                imports = [i.replace(" ", "") for i in imports]

                print("parents: ", parents)
                print("imports: ", imports)
                
                if len(parents) > 0:
                    file.write(f"from {c}{'.'.join(parents)} import {', '.join(imports)}\n")
                else:
                    file.write(f"from {c} import {', '.join(imports)}\n")
                
            else:
                file.write(line)

def add_dots_to_imports_in_folder(folder_path):
    for root, _, files in os.walk(folder_path):
        for file in files:
            if file.endswith('.py'):
                print("########\n", file)
                add_dots_to_imports(os.path.join(root, file))


# Clone the repo
clone_repo()
add_dots_to_imports_in_folder(f"./{project}/segmentation")


import os
import time
import yaml
import datetime
import numpy as np
from typing import List, Union
import torch
from torch.utils.data import DataLoader
import PIL
from torchvision.transforms.functional import to_tensor

from .model import FundusQualityModel, download_weights
from .default import MODELS_DIR, ENSEMBLE_MODELS
from pathlib import Path


def is_datetime_format(s):
    try:
        datetime.datetime.strptime(s, "%Y-%m-%d %H-%M-%S")
        return True
    except ValueError:
        return False


def any_to_tensor(image):
    """Converts an image to a torch tensor of shape (C, H, W)

    Args:
        image (str, np.ndarray, torch.Tensor): Image path, numpy array or torch tensor.

    Returns:
        torch.Tensor: Image as a torch tensor.
    """
    if isinstance(image, str):
        image = to_tensor(
            PIL.Image.open(image)
        )  # PIL Image with shape (H, W, C) -> (C, H, W)
    elif isinstance(image, np.ndarray):
        image = to_tensor(
            PIL.Image.fromarray(image)
        )  # numpy array with shape (H, W, C) -> (C, H, W)
    elif isinstance(image, torch.Tensor):
        image = image
    else:
        raise ValueError(
            "Image should be a string, numpy array or torch tensor but is of type",
            type(image),
        )
    return image


def get_ensemble(models_dir: str = MODELS_DIR, device: str = "cpu"):
    """Load the 10-model ensemble from the specified models or project directory.

    Args:
        device (str, optional): Device to use. Defaults to "cpu".
        models_dir (str, optional): Directory containing the "models" folder or its parent.
            Defaults to "./quality_prediction/models".

    Returns:
        ensemble: List of FundusQualityModel objects
    """
    models_dir = Path(models_dir)
    if "models" not in models_dir.parts:
        models_dir = models_dir / "models"
    models_dir.mkdir(parents=True, exist_ok=True)

    for e in ENSEMBLE_MODELS:
        if e not in [p.name for p in models_dir.iterdir() if p.is_dir()]:
            print("At least one model was not found.")
            download_weights()

    model_dirs = [
        p for p in models_dir.iterdir() if p.is_dir() and is_datetime_format(p.name)
    ]
    assert len(model_dirs) == len(
        ENSEMBLE_MODELS
    ), f"Expected 10 models, got {len(model_dirs)}. Did you download the models and place them into {models_dir}?"

    # Get configs and load models
    configs = []
    for p in model_dirs:
        with open(p / "config.yaml") as c:
            configs.append(yaml.safe_load(c))
            configs[-1]["device"] = device

    ensemble = []
    for ckpt, conf in zip(model_dirs, configs):
        model = FundusQualityModel(conf)
        model.load_checkpoint(str(ckpt))
        ensemble.append(model)

    return ensemble


def ensemble_predict(
    ensemble: List[FundusQualityModel],
    image: Union[list, str, torch.Tensor, np.ndarray],
    threshold: float = 0.5,
    print_result: bool = False,
):
    """Predicts the quality of an image or image batch (0: ungradable, 1: gradable)
        using an ensemble of models.

    Args:
        ensemble (list): List of FundusQualityModel objects.
        image (list, str, np.ndarray, torch.tensor): Image path(s) as a List[str] or an image as
            tensor or np.ndarray or a batch of images as a tensor.
        threshold (float, optional): Threshold for binary classification. Defaults to 0.5.

    Returns:
        ensemble_pred: Ensemble prediction(s) (confidence score between 0 and 1)
        binary_ensemble_pred: Ensemble predicted class(es), where 1 is good quality"""
    preds = []

    if isinstance(image, list) and isinstance(image[0], str):
        image = [any_to_tensor(i) for i in image]

    if (
        isinstance(image, (torch.Tensor, np.ndarray, list))
        and isinstance(image[0], (torch.Tensor, np.ndarray, PIL.Image.Image))
        and len(image[0].shape) == 3
    ):
        if not isinstance(image[0], torch.Tensor):
            image = [any_to_tensor(i) for i in image]
        # Do batch prediction
        n = len(image)
        preds = {i: [] for i in range(n)}
        for model in ensemble:
            pred = model.predict_from_batch(image, threshold=None, load_best=False)
            for i in range(n):
                preds[i].append(pred[i])
        for i in range(n):
            preds[i] = np.array(preds[i])
        ensemble_pred = np.array([preds[i].mean() for i in range(n)])
        binary_ensemble_pred = np.where(ensemble_pred >= threshold, 1, 0)

    elif isinstance(image, (list, str)) or isinstance(
        image, (torch.Tensor, np.ndarray)
    ):
        # Do single image prediction
        preds = []
        for model in ensemble:
            preds.append(
                model.predict_from_image(image, threshold=None, load_best=False)
            )
        preds = np.array(preds)

        ensemble_pred = preds.mean()
        binary_ensemble_pred = 1 if ensemble_pred >= threshold else 0

        if print_result:
            print(
                f"Ensemble confidence score: {ensemble_pred:.4f} \nEnsemble predicted class: {binary_ensemble_pred}, where 1 is good quality)"
            )

    else:
        raise ValueError(
            "Image(s) must be a list of paths, a path, or a tensor of the image or of a batch of images."
        )

    # Squeeze
    if isinstance(ensemble_pred, (list, np.ndarray)) and len(ensemble_pred) == 1:
        ensemble_pred = ensemble_pred[0]
    if (
        isinstance(binary_ensemble_pred, (list, np.ndarray))
        and len(binary_ensemble_pred) == 1
    ):
        binary_ensemble_pred = binary_ensemble_pred[0]

    return ensemble_pred, binary_ensemble_pred


def ensemble_predict_from_dataloader(
    ensemble: List[FundusQualityModel],
    dataloader: DataLoader,
    threshold: float = 0.5,
    print_result: bool = True,
):
    """Predicts the quality of images in a dataloader (0: ungradable, 1: gradable) using an ensemble of models.

    Args:
        ensemble (list): List of FundusQualityModel objects.
        dataloader (DataLoader): DataLoader containing the images.
        threshold (float, optional): Threshold for binary classification. Defaults to 0.5.
        print_result (bool, optional): Whether to print the result. Defaults to True.

    Returns:
        ensemble_preds: Ensemble prediction logits
        binary_ensemble_preds: Ensemble predicted classes, where 1 is good quality
        labels: True labels
    """
    labels = []
    preds = []
    for model in ensemble:
        pred, labels = model.predict_from_dataloader(dataloader, load_best=False)
        preds.append(pred)
    preds = np.array(preds)
    labels = np.array(labels)

    ensemble_preds = preds.mean(axis=0)
    binary_ensemble_preds = np.where(ensemble_preds >= threshold, 1, 0)

    assert (
        len(binary_ensemble_preds) == len(labels) == len(ensemble_preds)
    ), "Prediction and label lengths do not match."

    if print_result:
        print(
            f"Ensemble predicted logits: {ensemble_preds} \nEnsemble predicted classes: {binary_ensemble_preds}, where 1 is good quality)"
        )

    return ensemble_preds, binary_ensemble_preds, labels



import os
import yaml
from pprint import pprint
from types import SimpleNamespace
from fundus_image_toolbox.quality_prediction import (
    FundusQualityModel,
    FundusQualityLoader,
    MODELS_DIR,
)

from argparse import ArgumentParser
import sys


class Parser(ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n" % message)
        self.print_help()
        sys.exit(2)


if __name__ == "__main__":
    parser = Parser()
    parser.add_argument("--config", type=str, default=None)
    parser.add_argument("--train", type=bool, default=True)

    # Ignores these if config is passed
    parser.add_argument("--model_type", type=str, default="efficientnet-b4")
    parser.add_argument("--epochs", type=int, default=25)
    parser.add_argument("--img_size", type=int, default=350)
    parser.add_argument("--lr", type=float, default=1e-4)
    parser.add_argument("--device", type=str, default="cuda:0")
    parser.add_argument("--seed", type=int, default=12345)
    parser.add_argument(
        "--balance_datasets", type=bool, default=True
    )  # balances the ocurrence of samples from the different dataset origins without changing the total number of training samples
    parser.add_argument("--weight_decay", type=float, default=0.0)
    parser.add_argument(
        "--use_datasets", nargs="+", default=["drimdb", "deepdrid-isbi2020"]
    )  # "all", 'areds', 'registration', 'drimdb', 'deepdrid-isbi2020'
    parser.add_argument("--batch_size", type=int, default=8)

    args = parser.parse_args()

    if args.config is not None:
        print("Loading config, ignoring other arguments...")
        with open(args.config) as c:
            config = yaml.safe_load(c)
            pprint(config)
    else:
        config = {}
        config["model_type"] = args.model_type
        config["epochs"] = args.epochs
        config["img_size"] = args.img_size
        config["lr"] = args.lr
        config["weight_decay"] = args.weight_decay
        config["device"] = args.device
        config["seed"] = args.seed
        config["balance_datasets"] = args.balance_datasets
        config["use_datasets"] = args.use_datasets
        config["batch_size"] = args.batch_size

    config = SimpleNamespace(**config)

    TRAIN = args.train

    model = FundusQualityModel(config)

    train_dataloader, val_dataloader, test_dataloader = FundusQualityLoader(
        config
    ).get_dataloaders()

    if TRAIN:
        model.train(train_dataloader, val_dataloader)

    else:
        print("Looking for latest model...")
        # dirs = os.listdir(MODELS_DIR)
        dirs = MODELS_DIR.iterdir()
        ckpt = sorted(dirs)[-1]
        model.load_checkpoint(ckpt)

    result = model.evaluate(test_dataloader=test_dataloader)
    pprint(result)

    model.save_summary()

    # model.plot_grid(test_dataloader)



import torch
import scipy.stats as st
import torch.nn as nn
from torch.nn import functional as F
from PIL import Image
import numpy as np
import imgaug.augmenters as iaa
import os
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms

from ..common.common_util import pre_processing
from ..common.train_util import get_gaussian_kernel
import time

class RetinaDataset(Dataset):
    def __init__(self, data_path, split_file='eccv22_train.txt',
                 is_train=True, data_shape=(768, 768), auxiliary=None):
        self.data = []

        self.image_path = os.path.join(data_path, 'ImageData')
        self.label_path = os.path.join(data_path, 'Annotations')
        self.split_file = os.path.join(data_path, 'ImageSets', split_file)

        self.enhancement_sequential = iaa.Sequential([
            iaa.Multiply((1.0, 1.2)),  # change brightness, doesn't affect keypoints
            iaa.Sometimes(
                0.2,
                iaa.GaussianBlur(sigma=(0, 6))
            ),
            iaa.Sometimes(
                0.2,
                iaa.LinearContrast((0.75, 1.2))
            ),
        ], random_order=True)

        self.is_train = is_train
        self.model_image_height, self.model_image_width = data_shape[0], data_shape[1]

        with open(self.split_file) as f:
            files = f.readlines()

        for file in files:
            file = file.strip()
            try:
                image, label = file.split(', ')
                image = os.path.join(self.image_path, image)
                label = os.path.join(self.label_path, label)
                assert os.path.exists(image) and os.path.exists(label)
                self.data.append((image, label))
            except Exception:
                pass
        if auxiliary is not None and os.path.exists(auxiliary):
            print('-'*10+f"Load Lab {'train' if is_train else 'eval'} data and auxiliary data without labels"+'-'*10)
            auxiliaries_tmp = os.listdir(auxiliary)
            auxiliaries = []
            auxiliaries_tmp = [(os.path.join(auxiliary, a), None) for a in auxiliaries_tmp]
            for a, b in auxiliaries_tmp:
                if a.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif')):
                    auxiliaries.append((a, b))
            self.data = self.data + auxiliaries
        else:
            print('-' * 10 + f"Load Lab {'train' if is_train else 'eval'} data, and there is no auxiliary data" + '-' * 10)
        self.transforms = transforms.Compose([
            transforms.Resize((self.model_image_width, self.model_image_height)),
            transforms.ToTensor(),
        ])

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        input_with_label = False

        image_path, label_path = self.data[index]
        label_name = '[None]'

        image = Image.open(image_path).convert('RGB')
        image = np.asarray(image)
        image = image[:, :, 1]

        image = pre_processing(image)
        # if self.is_train:
        #     image = self.enhancement_sequential(image=image)
        image = Image.fromarray(image)
        image_tensor = self.transforms(image)

        if label_path is not None:
            label_name = os.path.split(label_path)[-1]
            keypoint_position = np.loadtxt(label_path)  # (2, n): (x, y).T
            keypoint_position[:, 0] *= image_tensor.shape[-1]
            keypoint_position[:, 1] *= image_tensor.shape[-2]

            tensor_position = torch.zeros([self.model_image_height, self.model_image_width])

            tensor_position[keypoint_position[:, 1], keypoint_position[:, 0]] = 1
            tensor_position = tensor_position.unsqueeze(0)
            input_with_label = True

            return image_tensor, input_with_label, tensor_position, label_name

        # without labels, only train descriptor
        tensor_position = torch.empty(image_tensor.shape)

        return image_tensor, input_with_label, tensor_position, label_name






from .model import FundusQualityModel, plot_quality
from .dataloader import FundusQualityLoader
from .ensemble_inference import get_ensemble as load_quality_ensemble
from .ensemble_inference import ensemble_predict as ensemble_predict_quality
from .ensemble_inference import ensemble_predict_from_dataloader as ensemble_predict_quality_from_dataloader
from .default import DATA_ROOT, MODELS_DIR



import numpy as np
import os
import yaml
from types import SimpleNamespace
from typing import List
from argparse import ArgumentParser
from pathlib import Path

from .dataset_multi import ODFoveaLoader
from .model_multi import ODFoveaModel
from .default import MODELS_DIR, DEFAULT_CSV_PATH, DEFAULT_MODEL, DEFAULT_CONFIG

import sys


class Parser(ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n" % message)
        self.print_help()
        sys.exit(2)


def get_ensemble(
    models_dir: str = MODELS_DIR,
    device: str = "cuda:0",
    return_test_dataloader: bool = False,
):
    ens = ["2024-05-07 11_13.05", "2024-05-07 11_13.13"]
    models = []
    checkpoints = []
    # for model_dir in [os.path.join(models_dir, x) for x in ens]:
    for model_dir in [Path(models_dir) / x for x in ens]:
        model, ckpt, test_dataloader = load_model(
            model_dir, device, return_test_dataloader=True
        )
        models.append(model)
        checkpoints.append(ckpt)

    if return_test_dataloader:
        return models, checkpoints, test_dataloader
    return models, checkpoints


def evaluate_ensemble(
    models: List[ODFoveaModel], checkpoints: List[str], test_dataloader: ODFoveaLoader
):
    test_loss, test_iou, test_dist, standardized_test_dist = [], [], [], []
    for model, ckpt in zip(models, checkpoints):
        loss, iou, dist, std_dist = model.evaluate(
            test_dataloader,
            checkpoint_path=ckpt,
            load_checkpoint=True,
            save_summary=False,
        )
        test_loss.append(loss)
        test_iou.append(iou)
        test_dist.append(dist)
        standardized_test_dist.append(std_dist)

    test_loss, test_iou, test_dist, standardized_test_dist = map(
        lambda x: sum(x) / len(x),
        [test_loss, test_iou, test_dist, standardized_test_dist],
    )

    return test_loss, test_iou, test_dist, standardized_test_dist


def predict_ensemble(models: List[ODFoveaModel], image_path: str):
    fovea_preds, od_preds = [], []
    for model in models:
        fovea_pred, od_pred = model.predict(image_path)
        fovea_preds.append(fovea_pred)
        od_preds.append(od_pred)

    # Mean of coordinates (x,y) of fovea and OD
    fovea_preds = np.mean(fovea_preds, axis=0)
    od_preds = np.mean(od_preds, axis=0)

    return fovea_preds, od_preds


def get_test_dataloader(config: SimpleNamespace):
    _, _, test_dataloader = ODFoveaLoader(config).get_dataloaders()
    return test_dataloader


def load_model(
    checkpoint_dir: str = "default",
    device: str = None,
    return_test_dataloader: bool = False,
):
    if checkpoint_dir == "default":
        checkpoint_dir = DEFAULT_MODEL
        config = DEFAULT_CONFIG
        checkpoint_dir = Path(checkpoint_dir)
    else:
        checkpoint_dir = Path(checkpoint_dir)
        # config = yaml.safe_load(open(os.path.join(checkpoint_dir, "config.yaml"), "r"))
        config = yaml.safe_load(open(checkpoint_dir / "config.yaml", "r"))
        config = SimpleNamespace(**config)

    if not MODELS_DIR in checkpoint_dir.parts:
        checkpoint_dir = Path(MODELS_DIR) / checkpoint_dir

    if config.csv_path is None:
        config.csv_path = DEFAULT_CSV_PATH
    if config.data_root is None:
        # config.data_root = "../../"
        config.data_root = Path(__file__).parent.parent

    config.device = device

    m_type = config.model_type
    model = ODFoveaModel(config)
    # model.checkpoint_path = os.path.join(checkpoint_dir, f"multi_{m_type}_best.pt")
    model.checkpoint_path = checkpoint_dir / f"multi_{m_type}_best.pt"
    model.load_checkpoint()

    # Load metric tracking data
    model.loss_tracking = {"train": [], "val": []}
    model.iou_tracking = {"train": [], "val": []}
    model.dist_tracking = {"train": [], "val": []}

    with open(checkpoint_dir / f"multi_{m_type}_train_loss.txt", "r") as f:
        for line in f:
            model.loss_tracking["train"].append(float(line.strip()))
    with open(checkpoint_dir / f"multi_{m_type}_val_loss.txt", "r") as f:
        for line in f:
            model.loss_tracking["val"].append(float(line.strip()))
    with open(checkpoint_dir / f"multi_{m_type}_train_iou.txt", "r") as f:
        for line in f:
            model.iou_tracking["train"].append(float(line.strip()))
    with open(checkpoint_dir / f"multi_{m_type}_val_iou.txt", "r") as f:
        for line in f:
            model.iou_tracking["val"].append(float(line.strip()))
    with open(checkpoint_dir / f"multi_{m_type}_train_dist.txt", "r") as f:
        for line in f:
            model.dist_tracking["train"].append(float(line.strip()))
    with open(checkpoint_dir / f"multi_{m_type}_val_dist.txt", "r") as f:
        for line in f:
            model.dist_tracking["val"].append(float(line.strip()))

    if return_test_dataloader:
        return model, model.checkpoint_path, get_test_dataloader(config)
    return model, model.checkpoint_path


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="If passed, other arguments are ignored and the model is trained using the config file",
    )
    parser.add_argument(
        "--model_type",
        type=str,
        default="efficientnet-b3",
        help="Type of model to train",
        choices=[
            "resnet18",
            "resnet34",
            "resnet50",
            "resnet101",
            "resnet152",
            "efficientnet-b0",
            "efficientnet-b1",
            "efficientnet-b2",
            "efficientnet-b3",
            "efficientnet-b4",
            "efficientnet-b5",
            "efficientnet-b6",
            "efficientnet-b7",
        ],
    )
    parser.add_argument(
        "--device", type=str, default="cuda:0", help="Device to train on"
    )
    parser.add_argument("--epochs", type=int, default=500, help="Number of epochs")
    parser.add_argument("--batch_size", type=int, default=8, help="Batch size")
    parser.add_argument("--lr", type=float, default=1e-4, help="Learning rate")
    parser.add_argument(
        "--img_size", type=int, default=350, help="Size of the input image"
    )
    parser.add_argument(
        "--testset_eval",
        type=bool,
        default=True,
        help="Evaluate on test set after training",
    )
    parser.add_argument(
        "--csv_path",
        type=str,
        default="../ADAM+IDRID+REFUGE_df.csv",
        help="Path to the csv file",
    )
    parser.add_argument(
        "--data_root",
        type=str,
        default="../",
        help="Root folder that the csv entries refer to",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=123,
        help="Seed for reproducibility. A different fixed seed is used for train-val-test split.",
    )
    config = parser.parse_args()

    if config.config is not None:
        config = yaml.safe_load(open(config.config, "r"))
        config = SimpleNamespace(**config)
    else:
        config = SimpleNamespace(**vars(config))

    print(f"Training {config.model_type} on {config.device}")
    if config.testset_eval:
        print("Evaluating on test set after training")

    train_dataloader, val_dataloader, test_dataloader = ODFoveaLoader(
        config
    ).get_dataloaders()
    model = ODFoveaModel(config)
    model.train(train_dataloader, val_dataloader)

    if config.testset_eval:
        model.evaluate(test_dataloader)

    model.plot_dist()
    model.plot_loss()
    model.plot_iou()



from datetime import datetime
import os
from types import SimpleNamespace
from typing import List, Union
from pathlib import Path
import requests

import yaml
import numpy as np
from tqdm import tqdm
from PIL import ImageDraw, Image
import matplotlib.pyplot as plt
import torch
import torch.nn as nn
from torchvision import models
from torchvision.ops import box_iou
from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor
from torchvision.transforms.functional import to_pil_image
from fundus_image_toolbox.utilities import ImageTorchUtils as Img

from .transforms_multi import ToPILImage
from .default import DEFAULT_MODEL, MODELS_DIR


def wget(link, target):
    # platform independent wget alternative
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36"
    }
    r = requests.get(link, headers=headers, stream=True)
    downloaded_file = open(target, "wb")

    for chunk in r.iter_content(chunk_size=8192):
        if chunk:
            downloaded_file.write(chunk)


class ODFoveaModel:
    def __init__(self, config: SimpleNamespace):
        self.config = config
        self.device = config.device
        self.model = self._get_model(config.model_type).to(self.device)
        self.timestamp = datetime.now().strftime("%Y-%m-%d %H_%M.%S")
        self.checkpoint_path = (
            MODELS_DIR / self.timestamp / f"multi_{self.config.model_type}_best.pt"
        )

        print(f"Initializing {self.config.model_type} on {self.device}")

        self.loss_func = nn.SmoothL1Loss(reduction="sum")
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.config.lr)

    def train(self, train_dataloader, val_dataloader):
        self.checkpoint_path.parent.mkdir(parents=True, exist_ok=True)
        if not (self.checkpoint_path.parent / "config.yaml").exists():
            yaml.safe_dump(
                vars(self.config),
                open(self.checkpoint_path.parent / "config.yaml", "w"),
            )

        best_epoch = 0
        self.loss_tracking = {"train": [], "val": []}
        self.iou_tracking = {"train": [], "val": []}
        self.dist_tracking = {"train": [], "val": []}
        self.best_loss = float("inf")

        for epoch in range(self.config.epochs):

            pbar = tqdm(total=len(train_dataloader))
            pbar.set_description(f"Epoch: {epoch+1} Training...")

            training_loss, training_iou, training_dist = self._train_val_step(
                train_dataloader, self.model, self.loss_func, self.optimizer, pbar
            )
            self.loss_tracking["train"].append(training_loss)
            self.iou_tracking["train"].append(training_iou)
            self.dist_tracking["train"].append(training_dist)

            pbar.set_description(f"Epoch: {epoch+1} Evaluating...")
            with torch.inference_mode():
                val_loss, val_iou, val_dist = self._train_val_step(
                    val_dataloader, self.model, self.loss_func, None, pbar
                )
                self.loss_tracking["val"].append(val_loss)
                self.iou_tracking["val"].append(val_iou)
                self.dist_tracking["val"].append(val_dist)

                if val_loss < self.best_loss:
                    best_epoch = epoch
                    torch.save(self.model.state_dict(), self.checkpoint_path)
                    self.best_loss = val_loss

            text = f"{epoch+1}/{self.config.epochs} - Train loss: {training_loss:.4}, IoU: {training_iou:.2}, Dist: {training_dist:.4} - Val loss: {val_loss:.4}, IoU: {val_iou:.2}, Dist: {val_dist:.4}"
            pbar.set_description(text)

        pbar.close()

        print(
            f'Best model at epoch {best_epoch} with loss {self.best_loss:.4} and IoU {self.iou_tracking["val"][best_epoch]:.2} and distance {self.dist_tracking["val"][best_epoch]:.4}'
        )
        print(f"Model saved at {self.checkpoint_path}")

        # Save summary to file
        p = self.checkpoint_path.parent
        with open(p / "summary.txt", "w") as f:
            f.write(f"Best epoch: {best_epoch}\n")
            f.write(f"Best loss: {self.best_loss:.4}\n")
            f.write(f'Best IoU: {self.iou_tracking["val"][best_epoch]:.2}\n')
            f.write(f'Best distance: {self.dist_tracking["val"][best_epoch]:.4}\n')

        # Save loss, IoU and distance tracking to files
        with open(p / f"multi_{self.config.model_type}_train_loss.txt", "w") as f:
            f.write("\n".join(map(str, self.loss_tracking["train"])))
        with open(p / f"multi_{self.config.model_type}_val_loss.txt", "w") as f:
            f.write("\n".join(map(str, self.loss_tracking["val"])))
        with open(p / f"multi_{self.config.model_type}_train_iou.txt", "w") as f:
            f.write("\n".join(map(str, self.iou_tracking["train"])))
        with open(p / f"multi_{self.config.model_type}_val_iou.txt", "w") as f:
            f.write("\n".join(map(str, self.iou_tracking["val"])))
        with open(p / f"multi_{self.config.model_type}_train_dist.txt", "w") as f:
            f.write("\n".join(map(str, self.dist_tracking["train"])))
        with open(p / f"multi_{self.config.model_type}_val_dist.txt", "w") as f:
            f.write("\n".join(map(str, self.dist_tracking["val"])))

        pbar.close()

    def evaluate(
        self,
        test_dataloader,
        checkpoint_path=None,
        load_checkpoint=True,
        save_summary=True,
    ):
        """Evaluate the model on the test set

        Args:
            test_dataloader (torch.utils.data.DataLoader): Test set dataloader
            checkpoint_path (str, optional): Path to the model checkpoint. If "latest", the latest
                model in "models/" will be used. Defaults to None."
            load_checkpoint (bool, optional): Whether to load the checkpoint. Defaults to True.
            save_summary (bool, optional): Whether to save the summary to a file. Defaults to True.
        """
        if checkpoint_path is not None:
            if checkpoint_path == "latest":
                checkpoint_path = sorted(MODELS_DIR.iterdir())[-1]
                checkpoint_path = (
                    checkpoint_path / f"multi_{self.config.model_type}_best.pt"
                )
            # if not os.path.exists(checkpoint_path):
            if not checkpoint_path.exists():
                raise FileNotFoundError(f"Checkpoint {checkpoint_path} not found")
            else:
                self.checkpoint_path = checkpoint_path

        if load_checkpoint:
            self.load_checkpoint()

        with torch.inference_mode():
            pbar = tqdm(total=len(test_dataloader))
            test_loss, test_iou, test_dist = self._train_val_step(
                test_dataloader, self.model, self.loss_func, None, pbar
            )
            standardized_test_dist = test_dist * self.config.img_size

            print(
                f"Test loss: {test_loss:.4}, IoU: {test_iou:.2}, Dist: {test_dist:.4}, Standardized dist: {standardized_test_dist:.4}"
            )

        if save_summary:
            with open(f"{self.checkpoint_path.parent}/summary.txt", "a") as f:
                f.write(f"Test loss: {test_loss:.4}\n")
                f.write(f"Test IoU: {test_iou:.2}\n")
                f.write(f"Normalized test distance: {test_dist:.4}\n")
                f.write(
                    f"Pixel distance in {self.config.img_size}px images: {standardized_test_dist:.4}\n"
                )

        return test_loss, test_iou, test_dist, standardized_test_dist

    def predict(
        self,
        images: Union[
            str,
            List[str],
            List[np.ndarray],
            List[torch.Tensor],
            List[Image.Image],
            np.ndarray,
            torch.Tensor,
            Image.Image,
        ],
        cpu_numpy=True,
    ):
        """Predict the fovea and optic disc locations in an image or image batch.

        Args:
            image: Torch image batch of same sized images -or- image paths to same sized images
                -or- image path -or- image (array, tensor or PIL image)

        Returns:
            (List of) Coordinates of the fovea and optic disc in the format (f_x, f_y, od_x, od_y)
        """
        self.model.eval()

        # To Tensor batch of images
        images = Img(images).to_batch().img

        orig_shapes = [(img.shape[1], img.shape[2]) for img in images]

        # Apply test-time augmentation
        transforms = Compose(
            [
                ToTensor(),
                Resize(self.config.img_size, antialias=True),
                CenterCrop(self.config.img_size),
            ]
        )
        images = torch.stack([transforms(to_pil_image(img)) for img in images]).to(
            self.device
        )

        with torch.inference_mode():
            outs = self.model(images)

            # Adjust labels to original image size
            for i in range(len(outs)):
                h, w = orig_shapes[i]
                f_x, f_y = outs[i][0], outs[i][1]
                od_x, od_y = outs[i][2], outs[i][3]
                outs[i] = torch.stack([f_x * w, f_y * h, od_x * w, od_y * h])

            if cpu_numpy:
                outs = [out.cpu().numpy() for out in outs]

            if len(outs) == 1:
                return outs[0]

        return outs

    def load_checkpoint(self):
        print(f"Loading model from {self.checkpoint_path}")
        if not self.checkpoint_path.exists():
            if DEFAULT_MODEL in self.checkpoint_path.__str__():
                print(f"Default model {DEFAULT_MODEL} not found, downloading...")
                if not (MODELS_DIR / DEFAULT_MODEL).exists():
                    self._download_weights()
            else:
                raise FileNotFoundError(f"Checkpoint {self.checkpoint_path} not found")

        self.model.load_state_dict(
            torch.load(
                self.checkpoint_path, map_location=self.device, weights_only=True
            )
        )

    def _get_model(self, type):
        if type == "resnet18":
            model = models.resnet18(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(512, 4)
        elif type == "resnet34":
            model = models.resnet34(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(512, 4)
        elif type == "resnet50":
            model = models.resnet50(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(2048, 4)
        elif type == "resnet101":
            model = models.resnet101(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(2048, 4)
        elif type == "resnet152":
            model = models.resnet152(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(2048, 4)
        elif type == "efficientnet-b0":
            model = models.efficientnet_b0(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1280, 4)
        elif type == "efficientnet-b1":
            model = models.efficientnet_b1(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1280, 4)
        elif type == "efficientnet-b2":
            model = models.efficientnet_b2(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1408, 4)
        elif type == "efficientnet-b3":
            model = models.efficientnet_b3(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1536, 4)
        elif type == "efficientnet-b4":
            model = models.efficientnet_b4(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1792, 4)
        elif type == "efficientnet-b5":
            model = models.efficientnet_b5(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(2048, 4)
        elif type == "efficientnet-b6":
            model = models.efficientnet_b6(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(2304, 4)
        elif type == "efficientnet-b7":
            model = models.efficientnet_b7(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(2560, 4)

        else:
            raise ValueError("Model type not supported")

        return model

    def _download_weights(
        self, url="https://zenodo.org/records/11174642/files/weights.tar.gz"
    ):
        MODELS_DIR.mkdir(parents=True, exist_ok=True)
        weights_path = (MODELS_DIR / "weights.tar.gz").__str__()
        wget(url, weights_path)
        print("Extracting weights...")
        os.system(f"tar -xzf {weights_path} -C {MODELS_DIR}")

        # Get Windows compatible folder names
        extracted_folders = sorted(MODELS_DIR.iterdir())
        for folder in extracted_folders:
            if folder.is_dir():
                new_folder_name = folder.name.replace(":", "_")
                folder.rename(folder.parent / new_folder_name)

        print("Removing tar file...")
        os.remove(weights_path)
        print("Done")

    def _train_val_step(self, dataloader, model, loss_func, optimizer=None, pbar=None):
        if optimizer is not None:
            model.train()
        else:
            model.eval()

        running_loss = 0
        running_iou = 0
        running_dist = 0

        for image_batch, f_label_batch, od_label_batch in dataloader:
            pbar.update(1)

            output_labels = model(image_batch)
            output_labels = (output_labels[:, :2], output_labels[:, 2:])
            label_batch = (f_label_batch, od_label_batch)

            loss_value, iou_metric_value = self._batch_loss(
                loss_func, output_labels, label_batch, optimizer
            )
            fovea_dist = self._distance_batch(output_labels[0], label_batch[0])
            od_dist = self._distance_batch(output_labels[1], label_batch[1])
            running_dist += (fovea_dist + od_dist) / 2

            running_loss += loss_value
            running_iou += iou_metric_value

        n = len(dataloader.dataset)
        return running_loss / n, running_iou / n, running_dist / n

    def _centroid_to_bbox(self, centroids, w=0.15, h=0.15):
        x0_y0 = centroids - torch.tensor([w / 2, h / 2]).to(self.device)
        x1_y1 = centroids + torch.tensor([w / 2, h / 2]).to(self.device)
        return torch.cat([x0_y0, x1_y1], dim=1)

    def _iou_batch(self, output_labels, target_labels):
        output_bbox = self._centroid_to_bbox(output_labels)
        target_bbox = self._centroid_to_bbox(target_labels)
        return torch.trace(box_iou(output_bbox, target_bbox)).item()

    def _distance_batch(self, output_labels, target_labels):
        # Sum of distances in batch
        dist = torch.sum(
            torch.sqrt(torch.sum((output_labels - target_labels) ** 2, dim=1))
        ).item()
        return dist

    def _batch_loss(self, loss_func, outputs, targets, optimizer=None):
        # Sums of losses and intersection over union/Jaccard index in batch
        loss = loss_func(outputs[0], targets[0])
        loss += loss_func(outputs[1], targets[1])

        with torch.no_grad():
            iou_metric = self._iou_batch(outputs[0], targets[0])
            iou_metric += self._iou_batch(outputs[1], targets[1])
            iou_metric /= 2
        if optimizer is not None:
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        return loss.item(), iou_metric

    def plot_loss(self):
        plt.plot(self.loss_tracking["train"], label="Train loss")
        plt.plot(self.loss_tracking["val"], label="Val loss")
        plt.legend()
        p = self.checkpoint_path.parent / f"multi_{self.config.model_type}_loss.png"
        plt.savefig(p)
        plt.show()
        plt.close()

    def plot_iou(self):
        plt.plot(self.iou_tracking["train"], label="Train IoU")
        plt.plot(self.iou_tracking["val"], label="Val IoU")
        plt.legend()
        p = self.checkpoint_path.parent / f"multi_{self.config.model_type}_iou.png"
        plt.savefig(p)
        plt.show()
        plt.close()

    def plot_dist(self):
        plt.plot(self.dist_tracking["train"], label="Train distance")
        plt.plot(self.dist_tracking["val"], label="Val distance")
        plt.legend()
        p = self.checkpoint_path.parent / f"multi_{self.config.model_type}_dist.png"
        plt.savefig(p)
        plt.show()
        plt.close()

    def _show_image_with_4_bounding_box(
        self, image, labels, target_labels, ax, w_h_bbox=(50, 50), thickness=2
    ):
        f_x, f_y, od_x, od_y = labels
        f_x_target, f_y_target, od_x_target, od_y_target = target_labels

        w, h = w_h_bbox

        image = image.copy()
        ImageDraw.Draw(image).rectangle(
            (
                (f_x_target - w // 2, f_y_target - h // 2),
                (f_x_target + w // 2, f_y_target + h // 2),
            ),
            outline="blue",
            width=thickness,
        )
        ImageDraw.Draw(image).rectangle(
            (
                (od_x_target - w // 2, od_y_target - h // 2),
                (od_x_target + w // 2, od_y_target + h // 2),
            ),
            outline="blue",
            width=thickness,
        )

        ImageDraw.Draw(image).rectangle(
            ((f_x - w // 2, f_y - h // 2), (f_x + w // 2, f_y + h // 2)),
            outline="green",
            width=thickness,
        )
        ImageDraw.Draw(image).rectangle(
            ((od_x - w // 2, od_y - h // 2), (od_x + w // 2, od_y + h // 2)),
            outline="green",
            width=thickness,
        )

        ax.imshow(image)

    def plot_grid(self, dataset, seed=12345):
        print("random choice with seed:", seed)
        print("pred: green")
        print("target: blue")

        self.model.eval()
        rng = np.random.default_rng(seed)  # create Generator object
        n_rows = 4  # number of rows in the image subplot
        n_cols = 8  # # number of cols in the image subplot
        indexes = rng.choice(range(len(dataset)), n_rows * n_cols, replace=False)

        fig, axs = plt.subplots(n_rows, n_cols, figsize=(20, 10))
        for i, ax in enumerate(axs.flatten()):
            image, f_labels, od_labels = dataset[indexes[i]]
            output_labels = self.model(image.unsqueeze(0))
            output_labels = (output_labels[:, :2], output_labels[:, 2:])

            labels = (f_labels.unsqueeze(0), od_labels.unsqueeze(0))

            iou1 = self._iou_batch(output_labels[0], labels[0])
            iou2 = self._iou_batch(output_labels[1], labels[1])

            iou = (iou1 + iou2) / 2
            _, labels = ToPILImage()((image, (f_labels, od_labels)))

            output_labels = (output_labels[0].squeeze(0), output_labels[1].squeeze(0))
            image, outputs = ToPILImage()((image, output_labels))

            self._show_image_with_4_bounding_box(image, outputs, labels, ax)
            ax.set_title(f"{iou:.2f}")
            ax.axis("off")
        p = self.checkpoint_path.parent / f"multi_{self.config.model_type}_grid.png"
        plt.savefig(p)
        plt.show()
        plt.close()


def plot_input(id, dataset):
    image, f_label, od_label = dataset[id]
    image, (f_label, od_label) = ToPILImage()((image, (f_label, od_label)))

    f_x, f_y = f_label
    od_x, od_y = od_label

    thickness = 2
    w, h = 50, 50

    image = image.copy()
    ImageDraw.Draw(image).rectangle(
        ((f_x - w // 2, f_y - h // 2), (f_x + w // 2, f_y + h // 2)),
        outline="blue",
        width=thickness,
    )
    ImageDraw.Draw(image).rectangle(
        ((od_x - w // 2, od_y - h // 2), (od_x + w // 2, od_y + h // 2)),
        outline="blue",
        width=thickness,
    )
    plt.imshow(image)


def plot_coordinates(
    fundus: Union[np.ndarray, List[np.ndarray]],
    coordinates: Union[np.ndarray, List[np.ndarray]],
    axs=None,
    return_axs=None,
):
    """Plot the fundus image with the predicted fovea and optic disc coordinates

    Args:
        fundus (Union[np.ndarray, List[np.ndarray]]): Fundus image or list of fundus images
        coordinates (Union[np.ndarray, List[np.ndarray]]): Predicted coordinates of the fovea and optic disc
        axs ([type], optional): Matplotlib axis. Defaults to None.
        return_axs ([type], optional): Whether to return the axis. Defaults to None.

    Returns:
        axs: (List of) Matplotlib axis, if return_axs is True
    """

    if not isinstance(fundus, list):
        fundus = [fundus]
        coordinates = [coordinates]

    max_ncols = 10
    num_images = len(fundus)
    num_cols = min(num_images, max_ncols)
    num_rows = int(np.ceil(num_images / num_cols))

    if axs is None:
        fig, axs = plt.subplots(
            num_rows, num_cols, figsize=(num_cols * 4, num_rows * 4)
        )

    if num_images == 1 and not isinstance(axs, (list, np.ndarray)):
        axs = [axs]

    for i, f in enumerate(fundus):
        fx, fy, ox, oy = coordinates[i]
        axs[i].imshow(f)
        axs[i].scatter(fx, fy, c="g", label="Predicted Fovea Center")
        axs[i].scatter(ox, oy, c="b", label="Predicted OD Center")
        axs[i].legend()
        axs[i].axis("off")

    if return_axs:
        if num_images == 1:
            axs = axs[0]
        return axs

    else:
        plt.show()
        plt.close()



import numpy as np
import torch
import os
import cv2
from typing import Union
from fundus_image_toolbox.utilities import ImageTorchUtils as Img
from PIL import Image

# Adapted from https://github.com/HzFu/EyeQ/blob/master/EyeQ_preprocess/fundus_prep.py


def imread(file_path, c=None):
    if c is None:
        im = cv2.imread(file_path)
    else:
        im = cv2.imread(file_path, c)

    if im is None:
        raise "Can not read image"

    if im.ndim == 3 and im.shape[2] == 3:
        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
    return im


def imwrite(file_path, image):
    if image.ndim == 3 and image.shape[2] == 3:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    cv2.imwrite(file_path, image)


def fold_dir(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)
    return folder


def get_mask_BZ(img):
    """Creates mask of fundus. The function does not explicitly create a circular mask. However,
    it does perform a series of operations that could result in a roughly circular mask if the
    region of interest in the input image is approximately circular and centrally located.

    How:
    1. The function first converts the input image to grayscale if it's not already, and then applies a binary threshold to create an initial mask. This separates the image into regions of interest (white) and background (black) based on pixel intensity.
    2. The function then creates a new mask that is the inverse of the original mask and performs flood fill operations from the corners. This helps to remove small artifacts or noise near the borders of the image.
    3. The original mask and the flood-filled mask are combined to create a new mask that preserves the regions of interest while removing noise.
    4. The function then applies morphological operations (erosion and dilation) using a rectangular structuring element. These operations help to smooth the mask and remove small regions or noise.

    The erosion operation shrinks the white regions (regions of interest), which can help to round off any irregularities. The dilation operation then expands the white regions back to their original size, but the expansion is uniform in all directions, which can help to maintain the rounded shape.

    So, while the function does not explicitly create a circular mask, the combination of thresholding, flood fill, and morphological operations can result in a roughly circular mask if the region of interest in the input image is approximately circular and centrally located.
    """
    if img.ndim == 3:
        gray_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    else:
        gray_img = img
    threshold = np.mean(gray_img) / 3 - 5
    _, mask = cv2.threshold(gray_img, max(0, threshold), 1, cv2.THRESH_BINARY)
    nn_mask = np.zeros((mask.shape[0] + 2, mask.shape[1] + 2), np.uint8)
    new_mask = (1 - mask).astype(np.uint8)
    _, new_mask, _, _ = cv2.floodFill(
        new_mask, nn_mask, (0, 0), (0), cv2.FLOODFILL_MASK_ONLY
    )
    _, new_mask, _, _ = cv2.floodFill(
        new_mask,
        nn_mask,
        (new_mask.shape[1] - 1, new_mask.shape[0] - 1),
        (0),
        cv2.FLOODFILL_MASK_ONLY,
    )
    mask = mask + new_mask
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (20, 20))
    mask = cv2.erode(mask, kernel)
    mask = cv2.dilate(mask, kernel)

    return mask


def _get_center_by_edge(mask):
    """Calculates the center coordinates of a circular region based on the edges of the mask. It uses the row and column sums of the mask to find the approximate center coordinates."""
    center = [0, 0]
    x = mask.sum(axis=1)
    center[0] = np.where(x > x.max() * 0.95)[0].mean()
    x = mask.sum(axis=0)
    center[1] = np.where(x > x.max() * 0.95)[0].mean()
    return center


def _get_radius_by_mask_center(mask, center):
    """Calculates the radius based on a given mask and center coordinates.
    In summary, this function takes a mask and center coordinates, performs morphological operations on the mask, calculates the distances of non-zero elements from the center, and determines the radius based on the distribution of distances.
    """
    mask = mask.astype(np.uint8)
    ksize = max(mask.shape[1] // 400 * 2 + 1, 3)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (ksize, ksize))
    mask = cv2.morphologyEx(mask, cv2.MORPH_GRADIENT, kernel)
    index = np.where(mask > 0)
    d_int = np.sqrt((index[0] - center[0]) ** 2 + (index[1] - center[1]) ** 2)
    b_count = np.bincount(np.ceil(d_int).astype(int))
    radius = np.where(b_count > b_count.max() * 0.995)[0].max()
    return radius


def _get_circle_by_center_bbox(shape, center, bbox, radius):
    "Gets mask from center and radius"
    center_mask = np.zeros(shape=shape).astype("uint8")
    center_tmp = (int(center[0]), int(center[1]))
    # Draw circle
    center_mask = cv2.circle(center_mask, center_tmp[::-1], int(radius), (1), -1)
    return center_mask


def get_mask(img):
    if img.ndim == 3:
        # raise 'image dim is not 3'
        g_img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    elif img.ndim == 2:
        g_img = img.copy()
    else:
        raise "image dim is not 1 or 3"
    h, w = g_img.shape
    shape = g_img.shape[0:2]
    g_img = cv2.resize(g_img, (0, 0), fx=0.5, fy=0.5)
    tg_img = cv2.normalize(g_img, None, 0, 255, cv2.NORM_MINMAX)
    tmp_mask = get_mask_BZ(tg_img)
    center = _get_center_by_edge(tmp_mask)
    radius = _get_radius_by_mask_center(tmp_mask, center)

    # Resize back
    center = [center[0] * 2, center[1] * 2]
    radius = int(radius * 2)
    s_h = max(0, int(center[0] - radius))
    s_w = max(0, int(center[1] - radius))
    bbox = (s_h, s_w, min(h - s_h, 2 * radius), min(w - s_w, 2 * radius))
    tmp_mask = _get_circle_by_center_bbox(shape, center, bbox, radius)
    return tmp_mask, bbox, center, radius


def get_center_radius(mask):
    center = _get_center_by_edge(mask)
    radius = _get_radius_by_mask_center(mask, center)

    return center, radius


def center_image(img, mask, center=None, radius=None):
    """Centers the circle in the image and resize st. circle edges touch image edges"""
    if center is None or radius is None:
        center, radius = get_center_radius(mask)
    h, w = img.shape[:2]

    # Scale image
    scale = w / (2 * radius)
    img = cv2.resize(img, (0, 0), fx=scale, fy=scale)
    mask = cv2.resize(mask, (0, 0), fx=scale, fy=scale)

    # Get center of image
    center_img = (h / 2, w / 2)

    # Get new center of circle
    center_circle = (center[0] * scale, center[1] * scale)

    # Get transform
    transform = np.array(center_img) - np.array(center_circle)

    # Apply transform
    M = np.float32([[1, 0, transform[1]], [0, 1, transform[0]]])
    img = cv2.warpAffine(img, M, (w, h))
    mask = cv2.warpAffine(mask, M, (w, h))

    return img, mask


def mask_image(img, mask):
    img[mask <= 0, ...] = 0
    return img


def remove_back_area(img, bbox=None, border=None):
    image = img
    if border is None:
        border = np.array(
            (
                bbox[0],
                bbox[0] + bbox[2],
                bbox[1],
                bbox[1] + bbox[3],
                img.shape[0],
                img.shape[1],
            ),
            dtype=int,
        )
    image = image[border[0] : border[1], border[2] : border[3], ...]
    return image, border


def supplemental_black_area(img, border=None):
    image = img
    if border is None:
        h, v = img.shape[0:2]
        max_l = max(h, v)
        if image.ndim > 2:
            image = np.zeros(shape=[max_l, max_l, img.shape[2]], dtype=img.dtype)
        else:
            image = np.zeros(shape=[max_l, max_l], dtype=img.dtype)
        border = (
            int(max_l / 2 - h / 2),
            int(max_l / 2 - h / 2) + h,
            int(max_l / 2 - v / 2),
            int(max_l / 2 - v / 2) + v,
            max_l,
        )
    else:
        max_l = border[4]
        if image.ndim > 2:
            image = np.zeros(shape=[max_l, max_l, img.shape[2]], dtype=img.dtype)
        else:
            image = np.zeros(shape=[max_l, max_l], dtype=img.dtype)
    image[border[0] : border[1], border[2] : border[3], ...] = img
    return image, border


def process_without_gb(img):
    borders = []
    mask, bbox, center, radius = get_mask(img)
    r_img = mask_image(img, mask)
    r_img, r_border = remove_back_area(r_img, bbox=bbox)
    mask, _ = remove_back_area(mask, border=r_border)
    borders.append(r_border)
    r_img, sup_border = supplemental_black_area(r_img)
    mask, _ = supplemental_black_area(mask, border=sup_border)
    borders.append(sup_border)
    return r_img, borders, (mask * 255).astype(np.uint8), center, radius


def process_path(image_path, save_path=None, size=None, suffix=".png"):
    """load img from path and circle crop"""
    dst_image = os.path.splitext(image_path.split("/")[-1])[0] + suffix
    dst_path = os.path.join(save_path, dst_image)

    img = imread(image_path)

    img, mask, center, radius = process_img(img, size=size)

    if save_path:
        imwrite(dst_path, img)

    return img, mask, center, radius


def process_img(img, size=None):
    try:
        img, borders, mask, center, radius = process_without_gb(img)

        if size:
            img = cv2.resize(img, size)
            mask = cv2.resize(mask, size)

    except:
        print("failed")

    img, mask = center_image(img, mask)
    center, radius = get_center_radius(mask)

    return img, mask, center, radius


def crop(
    img: Union[str, Image.Image, np.ndarray, torch.Tensor, list],
    size=(512, 512),
    return_all=False,
    to_numpy=True,
):
    """Fit a circle to the image (or images in a batch), center it and crop off the background.

    Args:
        img: Image path(s), RGB numpy array(s) or RGB torch tensor(s).
        size: Size of the output images. Default is (512, 512).
        return_all: If True, return the cropped images, masks, centers, and radii. Default is False.
        to_numpy: If True, return numpy arrays, else torch tensors. Default is True.

    Returns:
        Cropped images as numpy arrays or torch tensors or all the outputs if return_all is True.
    """
    input_is_a_batch = Img(img).is_batch_like()

    # Process images, batches of images, and paths to numpy batch
    img_batch = (
        Img(img).to_tensor(silent=True).set_channel_dim(-1).to_batch().to_numpy().img
    )

    # Process
    imgs, masks, centers, radii = [], [], [], []
    for img in img_batch:
        img, mask, center, radius = process_img(img, size=size)
        img = Img(img).to_numpy().img
        imgs.append(img)
        masks.append(mask)
        centers.append(center)
        radii.append(radius)

    imgs, masks, centers, radii = (
        np.array(imgs),
        np.array(masks),
        np.array(centers),
        np.array(radii),
    )

    # Convert to torch tensors
    if not to_numpy:
        imgs = Img(imgs).to_batch().img
        masks = Img(masks).to_batch().img
        centers = torch.tensor(centers)
        radii = torch.tensor(radii)

    # Return single element if input was single element
    if not input_is_a_batch:
        imgs, masks, centers, radii = imgs[0], masks[0], centers[0], radii[0]

    if return_all:
        return imgs, masks, centers, radii

    return imgs



import torch
import os

from torch.utils.data import DataLoader

from .common.train_util import train_model
from .dataset.retina_dataset import RetinaDataset
from .model.super_retina import SuperRetina
import torch.optim as optim
import yaml
from torch.optim import lr_scheduler
import warnings

warnings.filterwarnings("ignore")

config_path = "./config/train.yaml"

if os.path.exists(config_path):
    with open(config_path) as f:
        config = yaml.safe_load(f)
else:
    raise FileNotFoundError("Config File doesn't Exist")

assert "MODEL" in config
assert "PKE" in config
assert "DATASET" in config
assert "VALUE_MAP" in config
train_config = {
    **config["MODEL"],
    **config["PKE"],
    **config["DATASET"],
    **config["VALUE_MAP"],
}

batch_size = train_config["batch_size"]
num_epoch = train_config["num_epoch"]
device = train_config["device"]
device = torch.device(device if torch.cuda.is_available() else "cpu")

dataset_path = train_config["dataset_path"]
data_shape = (train_config["model_image_height"], train_config["model_image_width"])

train_split_file = train_config["train_split_file"]
val_split_file = train_config["val_split_file"]
auxiliary = train_config["auxiliary"]
train_set = RetinaDataset(
    dataset_path,
    split_file=train_split_file,
    is_train=True,
    data_shape=data_shape,
    auxiliary=auxiliary,
)
val_set = RetinaDataset(
    dataset_path, split_file=val_split_file, is_train=False, data_shape=data_shape
)

load_pre_trained_model = train_config["load_pre_trained_model"]
pretrained_path = train_config["pretrained_path"]

model = SuperRetina(train_config, device=device)
if load_pre_trained_model:
    if not os.path.exists(pretrained_path):
        raise Exception("Pretrained model doesn't exist")
    checkpoint = torch.load(pretrained_path, map_location=device, weights_only=True)
    model.load_state_dict(checkpoint["net"])

optimizer = optim.Adam(model.parameters(), lr=1e-4)

dataloaders = {
    "train": DataLoader(train_set, batch_size=batch_size, shuffle=True, num_workers=8),
    "val": DataLoader(val_set, batch_size=batch_size, shuffle=True, num_workers=8),
}

model = train_model(
    model,
    optimizer,
    dataloaders,
    device,
    num_epochs=num_epoch,
    train_config=train_config,
)



import torch.nn as nn
import torch
import torch.nn.functional as F


class DiceBCELoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(DiceBCELoss, self).__init__()

    def forward(self, inputs, targets, smooth=1):
        # comment out if your model contains a sigmoid or equivalent activation layer
        # inputs = F.sigmoid(inputs)

        # flatten label and prediction tensors
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        intersection = (inputs * targets).sum()
        dice_loss = 1 - (2. * intersection + smooth) / (inputs.sum() + targets.sum() + smooth)
        BCE = F.binary_cross_entropy(inputs, targets, reduction='mean')
        Dice_BCE = 0.01*BCE + dice_loss

        return Dice_BCE

class DiceLoss(nn.Module):
    def __init__(self, smooth=1e-3, p=2, reduction='mean'):
        super(DiceLoss, self).__init__()
        self.smooth = smooth
        self.p = p
        self.reduction = reduction

    def forward(self, predict, target):
        assert predict.shape[0] == target.shape[0], "predict & target batch size don't match"
        predict = predict.contiguous().view(-1)
        target = target.contiguous().view(-1)
        a = torch.mul(predict, target)
        b = predict.pow(self.p) + target.pow(self.p)

        num = 2 * torch.sum(a, dim=0) + self.smooth

        den = torch.sum(b, dim=0) + self.smooth

        loss = 1 - num / den

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss
        else:
            raise Exception('Unexpected reduction {}'.format(self.reduction))


import configparser

import numpy as np
from tqdm import tqdm

from .common.eval_util import compute_auc
from predictor import Predictor
import os
import cv2
import yaml

config_path = './config/test.yaml'
if os.path.exists(config_path):
    with open(config_path) as f:
        config = yaml.safe_load(f)
else:
    raise FileNotFoundError("Config File doesn't Exist")

Pred = Predictor(config)

data_path = './data/'  # Change the data_path according to your own setup
testset = 'FIRE'
use_matching_trick = config['PREDICT']['use_matching_trick']
gt_dir = os.path.join(data_path, testset, 'Ground Truth')
im_dir = os.path.join(data_path, testset, 'Images')

match_pairs = [x for x in os.listdir(gt_dir) if x.endswith('.txt')
               and not x.endswith('P37_1_2.txt')]

match_pairs.sort()
big_num = 1e6
good_nums_rate = []
image_num = 0

failed = 0
inaccurate = 0
mae = 0
mee = 0

# category: S, P, A, corresponding to Easy, Hard, Mod in paper
auc_record = dict([(category, []) for category in ['S', 'P', 'A']])

for pair_file in tqdm(match_pairs):
    gt_file = os.path.join(gt_dir, pair_file)
    file_name = pair_file.replace('.txt', '')

    category = file_name.split('_')[2][0]

    refer = file_name.split('_')[2] + '_' + file_name.split('_')[3]
    query = file_name.split('_')[2] + '_' + file_name.split('_')[4]

    query_im_path = os.path.join(im_dir, query + '.jpg')
    refer_im_path = os.path.join(im_dir, refer + '.jpg')
    H_m1, inliers_num_rate, query_image, _ = Pred.compute_homography(query_im_path, refer_im_path)
    H_m2 = None
    if use_matching_trick:
        if H_m1 is not None:
            h, w = Pred.image_height, Pred.image_width
            query_align_first = cv2.warpPerspective(query_image, H_m1, (w, h), borderMode=cv2.BORDER_CONSTANT,
                                              borderValue=(0))
            query_align_first = query_align_first.astype(float)
            query_align_first /= 255.
            H_m2, inliers_num_rate, _, _ = Pred.compute_homography(query_align_first, refer_im_path, query_is_image=True)

    good_nums_rate.append(inliers_num_rate)
    image_num += 1

    if inliers_num_rate < 1e-6:
        failed += 1
        avg_dist = big_num
    else:
        points_gd = np.loadtxt(gt_file)
        raw = np.zeros([len(points_gd), 2])
        dst = np.zeros([len(points_gd), 2])
        raw[:, 0] = points_gd[:, 2]
        raw[:, 1] = points_gd[:, 3]
        dst[:, 0] = points_gd[:, 0]
        dst[:, 1] = points_gd[:, 1]
        dst_pred = cv2.perspectiveTransform(raw.reshape(-1, 1, 2), H_m1)
        if H_m2 is not None:
            dst_pred = cv2.perspectiveTransform(dst_pred.reshape(-1, 1, 2), H_m2)

        dst_pred = dst_pred.squeeze()

        dis = (dst - dst_pred) ** 2
        dis = np.sqrt(dis[:, 0] + dis[:, 1])
        avg_dist = dis.mean()
        
        mae = dis.max()
        mee = np.median(dis)
        if mae > 50 or mee > 20:
            inaccurate += 1
    
    auc_record[category].append(avg_dist)
    

print('-'*40)
print(f"Failed:{'%.2f' % (100*failed/image_num)}%, Inaccurate:{'%.2f' % (100*inaccurate/image_num)}%, "
      f"Acceptable:{'%.2f' % (100*(image_num-inaccurate-failed)/image_num)}%")

print('-'*40)

auc = compute_auc(auc_record['S'], auc_record['P'], auc_record['A'])
print('S: %.3f, P: %.3f, A: %.3f, mAUC: %.3f' % (auc['s'], auc['p'], auc['a'], auc['mAUC']))



import os
from typing import List, Union
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

def exist(paths: Union[str, list]):
    """Check if a path or multiple paths exist."""
    if isinstance(paths, str):
        paths = [paths]
    for path in paths:
        if not os.path.exists(path):
            return False
    return True

def exists (paths: Union[str, list]):
    """Check if a path or multiple paths exist."""
    return exist(paths)

def flatten_one(l):
    """Flatten one level of nesting"""
    return [item for sublist in l for item in sublist]

def parse_list(arg: Union[list, str]):
    """Parse a list-like structures as a list."""
    if isinstance(arg, list):
        return arg
    elif (
        isinstance(arg, np.ndarray)
        or isinstance(arg, tuple)
        or isinstance(arg, set)
        or isinstance(arg, pd.Series)
    ):
        return list(arg)
    else:
        try:
            arg = str(arg)
            arg = arg.replace(" ", "").replace("[", "").replace("]", "")
            if "," in arg:
                return arg.split(",")
            else:
                # not a list
                return arg
        except Exception as e:
            raise f"Could not parse {arg} of type {type(arg)} as a list: " + str(e)
        
def on_slurm_job():
    """Check if the code is running on a SLURM job."""
    if "SLURM_JOB_ID" in os.environ:
        return True
    else:
        return False
    
def show(img: Union[np.ndarray, List[np.ndarray]]):
    if isinstance(img, list):
        for i, f in enumerate(img):
            plt.subplot(1, len(img), i+1)
            plt.imshow(f)
            plt.axis('off')
    else:
        plt.imshow(img)
        plt.axis('off')
    plt.show()

def print_type(img: Union[np.ndarray, List[np.ndarray]]):
    if isinstance(img, list):
        for i, f in enumerate(img):
            print_type(f)
    else:
        print(f"Type: {type(img)}")
        if isinstance(img, np.ndarray):
            print(f"Shape: {img.shape}")
            print(f"Type: {img.dtype}")
            print()


import shutil

import torch
from scipy.optimize import brentq
from scipy.interpolate import interp1d
from sklearn.metrics import roc_curve

import numpy as np
from predictor import Predictor
import os
import yaml
from tqdm import tqdm

config_path = "./config/test_VARIA.yaml"
if os.path.exists(config_path):
    with open(config_path) as f:
        config = yaml.safe_load(f)
else:
    raise FileNotFoundError("Config File doesn't Exist")

Pred = Predictor(config)

data_path = "./data/"  # Change the data_path according to your own setup
testset = "VARIA"
save_tmp_info = "VARIA_info"  # Save keypoints and descriptors
id_path = os.path.join(data_path, testset, "pair_index.txt")
im_dir = os.path.join(data_path, testset, "Images")

save_tmp_info = os.path.join(data_path, testset, save_tmp_info)

if os.path.exists(save_tmp_info):
    shutil.rmtree(save_tmp_info)
os.makedirs(save_tmp_info)

with open(id_path, "r") as f:
    pairs = f.readlines()

inliers = []
classes = []

print("Getting Predictions of All Images")
for img in tqdm(os.listdir(im_dir)):
    if img.endswith(".pgm"):
        image_path = os.path.join(im_dir, img)
        save_path = os.path.join(save_tmp_info, img.replace(".pgm", ".pt"))

        Pred.model_run_one_image(image_path, save_path)

print("Matching Pairs")
for pair in tqdm(pairs):
    pair = pair.strip()
    query, refer, is_accepted = pair.split(", ")

    query_im_path = os.path.join(im_dir, query)
    refer_im_path = os.path.join(im_dir, refer)

    query_info_path = os.path.join(save_tmp_info, query.replace(".pgm", ".pt"))
    refer_info_path = os.path.join(save_tmp_info, refer.replace(".pgm", ".pt"))

    query_info = torch.load(query_info_path, weights_only=True)
    refer_info = torch.load(refer_info_path, weights_only=True)

    _, inliers_num = Pred.homography_from_tensor(query_info, refer_info)

    inliers.append(inliers_num)
    classes.append(int(is_accepted))

inliers = np.array(inliers)
classes = np.array(classes)

fpr, tpr, threshold = roc_curve(classes, inliers)
eer = brentq(lambda x: 1.0 - x - interp1d(fpr, tpr)(x), 0.0, 1.0)
thresh = interp1d(fpr, threshold)(eer)

print("VARIA DATASET")
print("EER: %.2f%%, threshold: %d" % (eer * 100, thresh))



import os
from pathlib import Path
from types import SimpleNamespace
from pathlib import Path

import numpy as np
import pandas as pd
import torch

from PIL import Image
from torchvision import transforms
from fundus_image_toolbox.utilities import (
    multilevel_3way_split as split3,
    seed_everything,
)

from .transforms import get_transforms

# Data root. In there, folders "bad" and "good" with subfolders "drimdb" and "deepdrid-isbi2020"
# must exist.
from .default import DATA_ROOT

SPLIT_SEED = 12345


class FundusQualityDataset(torch.utils.data.Dataset):

    def __init__(
        self,
        config,
        DATA_ROOT=DATA_ROOT,
        split="train",
        augment=True,
        normalize=True,
        verbose=True,
        drimdb_dir=None,
        deepdrid_dir=None,
    ):
        if isinstance(config, dict):
            config = SimpleNamespace(**config)
        self.config = config

        self.data_root = DATA_ROOT
        self.split = split
        self.augment = augment
        seed_everything(self.config.seed, silent=True)
        self.verbose = verbose
        self.normalize = normalize
        self.data_dirs = {
            "drimdb": Path(drimdb_dir),
            "deepdrid-isbi2020": Path(deepdrid_dir),
        }

        self.data = self._load_data()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        img_name = self.data_root / self.data.iloc[idx]["image"]
        image = Image.open(img_name)
        label = self.data.iloc[idx]["label"]
        label = torch.tensor(label)
        origin = self.data.iloc[idx]["dataset"]

        if "drimdb" in origin:
            image = self._preprocess_drimdb(image)

        image = self._transform(image)

        return image.to(self.config.device), label.to(self.config.device)

    def _load_data(self) -> torch.utils.data.Dataset:
        datasets = (
            ["drimdb", "deepdrid-isbi2020"]
            if self.config.use_datasets == "all"
            else self.config.use_datasets
        )

        self._transfer_data()

        data = []
        for quality in ["bad", "good"]:
            for dataset in datasets:
                image_dir = self.data_root / quality / dataset
                for img_name in image_dir.iterdir():
                    if dataset == "deepdrid-isbi2020":
                        patient = img_name.split("_")[0]
                    else:
                        patient = len(data) - 10000
                    data.append(
                        [Path(quality) / dataset / img_name, quality, dataset, patient]
                    )
        df = pd.DataFrame(data, columns=["image", "quality", "dataset", "patient"])
        df["label"] = [0.0 if x == "bad" else 1.0 for x in df["quality"]]

        # Split the data individually by origin of dataset to yield resproducible splits independent of use_datasets
        train_df, val_df, test_df = (
            pd.DataFrame(columns=df.columns),
            pd.DataFrame(columns=df.columns),
            pd.DataFrame(columns=df.columns),
        )
        for dataset in datasets:
            _train_df, _val_df, _test_df = split3(
                df[df["dataset"] == dataset].reset_index(),
                [0.6, 0.2, 0.2],
                SPLIT_SEED,
                stratify_by="label",
                split_by="patient",
            )
            train_df = pd.concat([train_df, _train_df])
            val_df = pd.concat([val_df, _val_df])
            test_df = pd.concat([test_df, _test_df])

        if self.verbose:
            # Print label proportions, origins
            df = eval(self.split + "_df").copy(deep=True)
            print(f"{self.split} split:")
            print(df["label"].value_counts())
            print(df["dataset"].value_counts())
            print()
            del df

        if self.split == "train":
            return train_df
        elif self.split == "val":
            return val_df
        elif self.split == "test":
            return test_df

    def _transfer_data(self):
        """If DATA_ROOT does not yet contain the data, transfer it from the passed locations."""
        # Verify that desired folder structure exists
        for dataset in self.data_dirs:
            if not (self.data_root / "good" / dataset).exists():
                (self.data_root / "good" / dataset).mkdir(parents=True)
            if not (self.data_root / "bad" / dataset).exists():
                (self.data_root / "bad" / dataset).mkdir(parents=True)

        # Check if data is already present
        n_files = sum(
            [
                len(files)
                for r, d, files in os.walk(self.data_root)
                if any(f.endswith(".jpg") for f in files) and "iwebalbumfiles" not in r
            ]
        )
        if np.isclose(n_files, 2216, atol=50):
            return

        print("Copying data...")

        # Transfer DrimDB data
        drimdb_dir = self.data_dirs["drimdb"]
        for d in ["Bad", "Good", "Outlier"]:
            assert (
                Path(drimdb_dir) / d
            ).exists(), f"Directory {(Path(drimdb_dir) / d)} not found. Check your path to DrimDB data."
        for d in ["Bad", "Outlier"]:
            for f in (Path(drimdb_dir) / d).iterdir():
                os.system(
                    f"cp '{Path(drimdb_dir) / d / f}' '{self.data_root / 'bad' / 'drimdb' / f.name}'"
                )
        for f in (Path(drimdb_dir) / "Good").iterdir():
            os.system(
                f"cp '{Path(drimdb_dir) / 'Good' / f}' '{self.data_root / 'good' / 'drimdb' / f.name}'"
            )

        # Transfer DeepDrid data
        deepdrid_dir = self.data_dirs["deepdrid-isbi2020"]
        assert (
            Path(deepdrid_dir) / "regular_fundus_images"
        ).exists(), f"Directory {(Path(deepdrid_dir) / 'regular_fundus_images')} not found. Check your path to DeepDrid data."
        self._get_deepdrid()

        print("Done.")

    def _get_deepdrid(self, resize=True):
        def _resize_and_save(df, target_dirs, size=512):
            for i, row in df.iterrows():
                img = Image.open(row["image_path"])
                img = img.resize((size, size))
                label = row["Overall quality"]  # 1: good, 0: bad
                target_dir = target_dirs[label]
                suffix = ".jpg"
                img.save(target_dir / f"{row['image_id']}{suffix}")

        def _copy(df, target_dirs):
            for i, row in df.iterrows():
                img_path = row["image_path"]
                label = row["Overall quality"]
                # target_path = os.path.join(target_dirs[label], row["image_id"]+".jpg")
                target_path = target_dirs[label] / f"{row['image_id']}.jpg"
                os.system(f"cp '{img_path}' '{target_path}'")

        fundus_dir = Path(self.data_dirs["deepdrid-isbi2020"]) / "regular_fundus_images"
        train_data = fundus_dir / "regular-fundus-training"
        valid_data = fundus_dir / "regular-fundus-validation"
        test_data = fundus_dir / "Online-Challenge1&2-Evaluation"

        train_df = pd.read_csv(train_data / "regular-fundus-training.csv")
        val_df = pd.read_csv(valid_data / "regular-fundus-validation.csv")
        test_df = pd.read_excel(test_data / "Challenge2_labels.xlsx")

        train_df["image_path"] = train_df["image_id"].apply(
            lambda x: train_data / "Images" / x.split("_")[0] / f"{x}.jpg"
        )
        val_df["image_path"] = val_df["image_id"].apply(
            lambda x: valid_data / "Images" / x.split("_")[0] / f"{x}.jpg"
        )
        test_df["image_path"] = test_df["image_id"].apply(
            lambda x: test_data / "Images" / x.split("_")[0] / f"{x}.jpg"
        )

        df = pd.concat([train_df, val_df, test_df], ignore_index=True)

        target_dir_good = self.data_root / "good" / "deepdrid-isbi2020"
        target_dir_bad = self.data_root / "bad" / "deepdrid-isbi2020"

        target_dirs = {1: target_dir_good, 0: target_dir_bad}

        if resize:
            _resize_and_save(df, target_dirs)
        else:
            _copy(df, target_dirs)

    def _transform(self, image):
        split = self.split if self.augment else "test"
        return get_transforms(
            img_size=self.config.img_size, split=split, normalize=self.normalize
        )(image)

    def _preprocess_drimdb(self, image: Image.Image):
        """Preprocess DRIMDB images before applying transforms"""

        # 1. : Cut off the top 5% of height to remove the text
        cut = int(image.height * 0.05)
        image = image.crop((0, cut, image.width, image.height))

        # 2. : Resize to square by adding black padding at top and bottom
        size = max(image.size)
        new_image = Image.new("RGB", (size, size), (0, 0, 0))
        new_image.paste(image, (0, (size - image.size[1]) // 2))
        out = new_image

        return out



# Adapted from Huang et al., 2023
# https://github.com/YijinHuang/pytorch-classification/blob/master/data/transforms.py

from typing import Any, Dict
from torchvision import transforms

# ImageNet pixel mean and std for each channel
MEAN, STD = [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]

def get_transforms(img_size: int = 512, split: str = "train", normalize=True, mean=MEAN, std=STD):
    """Data transformations for fundus images, inspired by Huang et al. (2023).

    Args:
        img_size (int): image size (width -or- height)
        split (str): train, val, validation, or test

    Returns:
        callable transforms.Compose() object
    """

    if split != "train":
        transformations = [
            transforms.ToTensor(),
            transforms.Resize(img_size, antialias=True),
            transforms.CenterCrop(img_size),
        ]

        if not normalize:
            return transforms.Compose(transformations)
        return transforms.Compose(transformations+get_normalization(mean,std))

    aug_args: Dict[str, Any] = {
        "random_crop": {"scale": [0.85, 1.15], "ratio": [0.75, 1.25], "p": 1.0},
        "flip": {"p": 0.5},
        "color_jitter": {
            "p": 0.5,
            "brightness": 0.2,
            "contrast": 0.2,
            "saturation": 0,
            "hue": 0,
        },
        "rotation": {"p": 1.0, "degrees": [-180, 180]},
    }

    operations = {
        "random_crop": random_apply(
            transforms.RandomResizedCrop(
                size=(img_size, img_size),
                scale=aug_args["random_crop"]["scale"],
                ratio=aug_args["random_crop"]["ratio"],
                antialias=True,
            ),
            p=aug_args["random_crop"]["p"],
        ),
        "horizontal_flip": transforms.RandomHorizontalFlip(p=aug_args["flip"]["p"]),
        "vertical_flip": transforms.RandomVerticalFlip(p=aug_args["flip"]["p"]),
        "color_jitter": random_apply(
            transforms.ColorJitter(
                brightness=aug_args["color_jitter"]["brightness"],
                contrast=aug_args["color_jitter"]["contrast"],
                saturation=aug_args["color_jitter"]["saturation"],
                hue=aug_args["color_jitter"]["hue"],
            ),
            p=aug_args["color_jitter"]["p"],
        ),
        "rotation": random_apply(
            transforms.RandomRotation(degrees=aug_args["rotation"]["degrees"], fill=0),
            p=aug_args["rotation"]["p"],
        ),
    }

    augmentations = [transforms.ToTensor()]
    for operation in operations.items():
        augmentations.append(operation[1])

    if not normalize:
        return transforms.Compose(augmentations)
    return transforms.Compose(augmentations+get_normalization(mean,std))


def random_apply(op, p):
    """Randomly apply a transformation"""
    return transforms.RandomApply([op], p=p)


def get_normalization(mean=MEAN, std=STD):
    """Add normalization transformation using dataset wide pixel mean and std.
    
    Args:
        mean: list of mean values for each channel. Defaults to ImageNet mean.
        std: list of std values for each channel. Defaults to ImageNet std.
    
    Returns:
        list: transforms.Normalize wrapped by a list"""

    normalization = transforms.Normalize(
        mean, std
    )
    return [normalization]

def get_unnormalization(mean=MEAN, std=STD):
    """Add unnormalization transformation using original dataset wide pixel mean and std.
    
    Args:
        mean: list of mean values for each channel. Defaults to ImageNet mean.
        std: list of std values for each channel. Defaults to ImageNet std.
    
    Returns:
        transforms.Normalize: unnormalization transformation
    """
    new_mean = [-m/s for m,s in zip(mean,std)]
    new_std = [1/s for s in std]

    return get_normalization(new_mean, new_std)[0]



from torch.nn import functional as F
import torch
import matplotlib.pyplot as plt

from torchvision import transforms
import numpy as np
from PIL import Image
import cv2
import torch.nn as nn
import scipy.stats as st


def remove_borders(keypoints, scores, border: int, height: int, width: int):
    """ Removes keypoints too close to the border """
    mask_h = (keypoints[:, 0] >= border) & (keypoints[:, 0] < (height - border))
    mask_w = (keypoints[:, 1] >= border) & (keypoints[:, 1] < (width - border))
    mask = mask_h & mask_w
    return keypoints[mask], scores[mask]


def simple_nms(scores, nms_radius: int):
    """ Fast Non-maximum suppression to remove nearby points """
    assert (nms_radius >= 0)

    size = nms_radius * 2 + 1
    avg_size = 2
    def max_pool(x):
        return torch.nn.functional.max_pool2d(
            x, kernel_size=size, stride=1, padding=nms_radius)

    zeros = torch.zeros_like(scores)
    # max_map = max_pool(scores)

    max_mask = scores == max_pool(scores)
    max_mask_ = torch.rand(max_mask.shape).to(max_mask.device) / 10
    max_mask_[~max_mask] = 0
    mask = ((max_mask_ == max_pool(max_mask_)) & (max_mask_ > 0))

    return torch.where(mask, scores, zeros)


def pre_processing(data):
    """ Enhance retinal images """
    train_imgs = datasets_normalized(data)
    train_imgs = clahe_equalized(train_imgs)
    train_imgs = adjust_gamma(train_imgs, 1.2)

    train_imgs = train_imgs / 255.

    return train_imgs.astype(np.float32)


def rgb2gray(rgb):
    """ Convert RGB image to gray image """
    r, g, b = rgb.split()
    return g


def clahe_equalized(images):
    clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
    images_equalized = np.empty(images.shape)
    images_equalized[:, :] = clahe.apply(np.array(images[:, :],
                                                  dtype=np.uint8))

    return images_equalized


def datasets_normalized(images):
    # images_normalized = np.empty(images.shape)
    images_std = np.std(images)
    images_mean = np.mean(images)
    images_normalized = (images - images_mean) / (images_std + 1e-6)
    minv = np.min(images_normalized)
    images_normalized = ((images_normalized - minv) /
                         (np.max(images_normalized) - minv)) * 255

    return images_normalized


def adjust_gamma(images, gamma=1.0):
    invGamma = 1.0 / gamma
    table = np.array([((i / 255.0) ** invGamma) * 255
                      for i in np.arange(0, 256)]).astype("uint8")
    new_images = np.empty(images.shape)
    new_images[:, :] = cv2.LUT(np.array(images[:, :],
                                        dtype=np.uint8), table)

    return new_images


def nms(detector_pred, nms_thresh=0.1, nms_size=10, detector_label=None, mask=False):
    """ Apply NMS on predictions, if mask, then remove geo_points that appearing in labels """
    detector_pred = detector_pred.clone().detach()

    B, _, h, w = detector_pred.shape

    # if mask:
    #     assert detector_label is not None
    #     detector_pred[detector_pred < nms_thresh] = 0
    #     label_mask = detector_label
    #
    #     # more area
    #
    #     detector_label = detector_label.long().cpu().numpy()
    #     detector_label = detector_label.astype(np.uint8)
    #     kernel = np.ones((3, 3), np.uint8)
    #     label_mask = np.array([cv2.dilate(detector_label[s, 0], kernel, iterations=1)
    #                            for s in range(len(detector_label))])
    #     label_mask = torch.from_numpy(label_mask).unsqueeze(1)
    #     detector_pred[label_mask > 1e-6] = 0

    scores = simple_nms(detector_pred, nms_size)

    scores = scores.reshape(B, h, w)

    points = [
        torch.nonzero(s > nms_thresh)
        for s in scores]

    scores = [s[tuple(k.t())] for s, k in zip(scores, points)]

    points, scores = list(zip(*[
        remove_borders(k, s, 8, h, w)
        for k, s in zip(points, scores)]))
    points = [torch.flip(k, [1]).long() for k in points]

    return points


def sample_keypoint_desc(keypoints, descriptors, s: int = 8):
    """ Interpolate descriptors at keypoint locations """
    b, c, h, w = descriptors.shape
    keypoints = keypoints.clone().float()

    keypoints /= torch.tensor([(w * s - 1), (h * s - 1)]).to(keypoints)[None]
    keypoints = keypoints * 2 - 1  # normalize to (-1, 1)

    args = {'align_corners': True} if int(torch.__version__[2]) > 2 else {}
    descriptors = torch.nn.functional.grid_sample(
        descriptors, keypoints.view(b, 1, -1, 2), mode='bilinear', **args)

    descriptors = torch.nn.functional.normalize(
        descriptors.reshape(b, c, -1), p=2, dim=1)
    return descriptors


def sample_descriptors(detector_pred, descriptor_pred, affine_descriptor_pred, grid_inverse,
                       nms_size=10, nms_thresh=0.1, scale=8, affine_detector_pred=None):
    """
    sample descriptors based on keypoints
    :param affine_descriptor_pred:
    :param descriptor_pred:
    :param detector_pred:
    :param grid_inverse: used for inverse transformation of affine
    :param nms_size
    :param nms_thresh
    :param scale: down sampling size of detector
    :return: sampled descriptors
    """
    B, _, h, w = detector_pred.shape
    keypoints = nms(detector_pred, nms_size=nms_size, nms_thresh=nms_thresh)

    affine_keypoints = [(grid_inverse[s, k[:, 1].long(), k[:, 0].long()]) for s, k in
                        enumerate(keypoints)]

    kp = []
    affine_kp = []
    for s, k in enumerate(affine_keypoints):
        idx = (k[:, 0] < 1) & (k[:, 0] > -1) & (k[:, 1] < 1) & (
                k[:, 1] > -1)
        kp.append(keypoints[s][idx])
        ak = k[idx]
        ak[:, 0] = (ak[:, 0] + 1) / 2 * (w - 1)
        ak[:, 1] = (ak[:, 1] + 1) / 2 * (h - 1)
        affine_kp.append(ak)

    descriptors = [sample_keypoint_desc(k[None], d[None], s=scale)[0]
                   for k, d in zip(kp, descriptor_pred)]
    affine_descriptors = [sample_keypoint_desc(k[None], d[None], s=scale)[0]
                          for k, d in zip(affine_kp, affine_descriptor_pred)]
    return descriptors, affine_descriptors, keypoints



from argparse import ArgumentParser
import sys
import pickle
import os
import numpy as np
import pandas as pd
import torch
import cv2
import matplotlib.pyplot as plt
from typing import Union, Tuple, List
from PIL import Image
from .default import MODELS_DIR
from fundus_image_toolbox.utilities import ImageTorchUtils as Img
from fundus_image_toolbox.utilities import seed_everything

# Clone the repository if not present
try:
    from .segmentation.utils.notebook_utils import clahe_equalized, get_ensemble
    from .segmentation.utils.model_definition import FR_UNet
except ImportError:
    from .clone import clone_repo

    this_dir = os.path.dirname(os.path.abspath(__file__))
    clone_repo(target_dir=os.path.join(this_dir, "segmentation").__str__())
    from .segmentation.utils.notebook_utils import clahe_equalized, get_ensemble
    from .segmentation.utils.model_definition import FR_UNet


class Parser(ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n" % message)
        self.print_help()
        sys.exit(2)


def plot_masks(images, masks):
    """Plot images and masks side by side.

    Args:
        images (list, str): Path(s) to image(s) or image(s) themselves (np.ndarray, PIL.Image, torch.Tensor)
        masks (list, np.ndarray): Mask(s)

    Returns:
        None
    """
    images = Img(images).to_batch().to_numpy().img
    masks = [masks] if len(np.array(masks).shape) == 2 else masks

    if len(images) != len(masks):
        raise ValueError("Number of images and masks must be the same.")

    fig, ax = plt.subplots(len(images), 2, figsize=(10, 5 * len(images)))
    if len(images) == 1:
        ax = np.expand_dims(ax, axis=0)
    for i in range(len(images)):
        ax[i, 0].imshow(images[i])
        ax[i, 1].imshow(masks[i], cmap="gray")
    plt.show()


def save_masks(x_paths, masks, path):
    """Save masks to a pickle file.

    Args:
        x_paths (list, str): Path(s) to image(s)
        masks (list, np.ndarray): Mask(s)
        path (str): Directorry to save the masks in

    Returns:
        None
    """
    if isinstance(x_paths, str):
        x_paths = [x_paths]
        masks = [masks]
    masks_flattened = [mask.flatten() for mask in masks]
    basenames = [os.path.basename(x) for x in x_paths]
    df = pd.DataFrame(
        {
            "image": basenames,
            "mask": masks_flattened,
            "shape": [mask.shape for mask in masks],
        }
    )
    p = os.path.join(path, "masks.pkl")
    os.makedirs(os.path.dirname(p), exist_ok=True)

    if os.path.exists(p):
        df_ = pickle.load(open(p, "rb"))
        df_ = df_[~df_["image"].isin(df["image"])]
        df = pd.concat([df_, df], axis=0)

    pickle.dump(df, open(p, "wb"))


def load_ensemble(device: str = "cuda:0"):
    models_paths = [
        os.path.join(MODELS_DIR, f)
        for f in os.listdir(MODELS_DIR)
        if f.endswith(".pth")
    ]
    ensemble_models = get_ensemble(models_paths, dropout=False, device=device)
    return ensemble_models


def predict(
    ensemble_models: list,
    images: Union[
        List[str],
        str,
        List[Image.Image],
        Image.Image,
        List[np.ndarray],
        np.ndarray,
        List[torch.Tensor],
        torch.Tensor,
    ],
    device: str = None,
    size: Tuple[int, int] = (512, 512),
    save_to: str = None,
    plot=False,
    threshold=0.5,
):
    """Predict vessel mask for a single image or a batch of images.

    Args:
        ensemble_models (list): List of models to ensemble.
        images (list, str, Image.Image, np.ndarray, torch.Tensor): List of paths to images, or a
            single path or a batch of image objects or a single image.
        device (str): Device to run inference on. If None, the device the model is on is used.
        size (tuple): Size of the output masks.
        save_to (str): Path to save the masks, optional.
        plot (bool): Whether to plot the masks.
        threshold (float): Threshold for binarization.

    Returns:
        np.ndarray: Predicted mask(s)
    """
    if device is None:
        device = str(next(ensemble_models[0].parameters()).device)

    seed_everything(23, silent=True)

    images = Img(images).to_batch().img
    images = [
        Img(image).to_tensor(from_cspace="rgb", to_cspace="bgr").to_numpy().img
        for image in images
    ]

    pred_masks = [[] for _ in range(len(ensemble_models))]

    for j, model in enumerate(ensemble_models):
        model = model.to(device)

        processed_images = []
        original_sizes = []
        for i, x in enumerate(images):
            height, width = x.shape[:2]
            original_sizes.append((width, height))

            if size[0] is None:
                size = (int(size[1] * width / height), size[1])
            elif size[1] is None:
                size = (size[0], int(size[0] * height / width))

            x = clahe_equalized(x)  # BGR > RGB
            x = cv2.resize(x, size)
            x = Img(x).to_tensor().img.to(device)

            processed_images.append(x)

        processed_images = torch.stack(processed_images)  # (n, c, h, w)

        with torch.no_grad():
            pred_logit = model(processed_images)
            pred = torch.sigmoid(pred_logit)  # (n, 1, h, w)
            pred_mask = np.array(pred.cpu() > threshold, dtype=int)
            pred_mask = pred_mask.squeeze()  # (n, h, w) or (h, w)
            if len(pred_mask.shape) == 2:
                pred_mask = np.expand_dims(pred_mask, axis=0)  # (n=1, h, w)

            pred_masks[j].append(pred_mask)

    masks = np.mean(pred_masks, axis=0)  # (1, n, h, w)
    masks = masks.squeeze()  # (n, h, w)
    if len(masks.shape) == 2:
        masks = np.expand_dims(masks, axis=0)  # (n=1, h, w)

    # Strech mask to input image size
    resized_masks = []
    for i, m in enumerate(masks):
        m = cv2.resize(m, original_sizes[i])
        resized_masks.append(m)
    masks = np.array(resized_masks)

    if plot:
        plot_masks(images, masks)

    if save_to is not None:
        save_masks(images, masks, save_to)

    if len(masks) == 1:
        return masks[0]

    return masks


def load_masks_from_filenames(filenames, masks_dir: str = None):
    """Load masks from a pickle file.

    Args:
        filenames (list, str): Path(s) to image(s)
        masks_dir (str): Directory to load the masks from

    Returns:
        list or ndarray: Mask(s)
    """
    if isinstance(filenames, str):
        filenames = [filenames]

    p = os.path.join(masks_dir, "masks.pkl")
    masks = pickle.load(open(p, "rb"))
    out = []

    for filename in filenames:
        row = masks.loc[masks["image"] == os.path.basename(filename)]
        mask = row["mask"].values[0]
        mask = np.array(mask)
        mask = mask.reshape(row["shape"].values[0])
        out.append(mask)

    if len(out) == 1:
        return out[0]
    return out


if __name__ == "__main__":
    parser = Parser()
    parser.add_argument(
        "--paths",
        nargs="+",
        type=str,
        default=None,
        help="Paths to images to be segmented.",
    )
    parser.add_argument(
        "--device", type=str, default="cuda:0", help="Device to run inference on."
    )
    parser.add_argument(
        "--size", type=int, default=512, help="Size of the output masks."
    )
    parser.add_argument(
        "--save_to", type=str, default=None, help="Path to save the masks, optional."
    )
    parser.add_argument(
        "--plot", type=bool, default=False, help="Whether to plot the masks."
    )
    parser.add_argument(
        "--threshold", type=float, default=0.5, help="Threshold for binarization."
    )
    args = parser.parse_args()

    ensemble_models = load_ensemble(args.device)
    predict(
        ensemble_models,
        args.paths,
        args.device,
        args.size,
        args.save_to,
        args.plot,
        args.threshold,
    )



from types import SimpleNamespace
import yaml
from fundus_image_toolbox.fovea_od_localization import (
    ODFoveaLoader,
    ODFoveaModel,
    Parser,
)

if __name__ == "__main__":
    parser = Parser()
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="If passed, other arguments are ignored and the model is trained using the config file",
    )
    parser.add_argument(
        "--model_type",
        type=str,
        default="efficientnet-b3",
        help="Type of model to train",
        choices=[
            "resnet18",
            "resnet34",
            "resnet50",
            "resnet101",
            "resnet152",
            "efficientnet-b0",
            "efficientnet-b1",
            "efficientnet-b2",
            "efficientnet-b3",
            "efficientnet-b4",
            "efficientnet-b5",
            "efficientnet-b6",
            "efficientnet-b7",
        ],
    )
    parser.add_argument(
        "--device", type=str, default="cuda:0", help="Device to train on"
    )
    parser.add_argument("--epochs", type=int, default=500, help="Number of epochs")
    parser.add_argument("--batch_size", type=int, default=8, help="Batch size")
    parser.add_argument("--lr", type=float, default=1e-4, help="Learning rate")
    parser.add_argument(
        "--img_size", type=int, default=350, help="Size of the input image"
    )
    parser.add_argument(
        "--testset_eval",
        type=bool,
        default=True,
        help="Evaluate on test set after training",
    )
    parser.add_argument(
        "--csv_path",
        type=str,
        default="../ADAM+IDRID+REFUGE_df.csv",
        help="Path to the csv file",
    )
    parser.add_argument(
        "--data_root",
        type=str,
        default="../",
        help="Root folder that the csv entries refer to",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=123,
        help="Seed for reproducibility. A different fixed seed is used for train-val-test split.",
    )
    config = parser.parse_args()

    if config.config is not None:
        config = yaml.safe_load(open(config.config, "r"))
        config = SimpleNamespace(**config)
    else:
        config = SimpleNamespace(**vars(config))

    print(f"Training {config.model_type} on {config.device}")
    if config.testset_eval:
        print("Evaluating on test set after training")

    train_dataloader, val_dataloader, test_dataloader = ODFoveaLoader(
        config
    ).get_dataloaders()
    model = ODFoveaModel(config)
    model.train(train_dataloader, val_dataloader)

    if config.testset_eval:
        model.evaluate(test_dataloader)

    model.plot_dist()
    model.plot_loss()
    model.plot_iou()



from .inference import plot_masks, save_masks, load_masks_from_filenames, FR_UNet
from .inference import load_ensemble as load_segmentation_ensemble
from .inference import predict as ensemble_predict_segmentation